#! /usr/bin/env python

import openturns as ot
import openturns.testing as ott
import numpy as np

ot.TESTPREAMBLE()


sample = ot.Normal(3).getSample(1000)
tree = ot.KDTree(sample)
print("tree=", tree)

test = ot.Normal(3).getSample(100)

sample_np = np.array(sample)
test_np = np.array(test)


def nearest_debug(x):
    a = np.sum(np.square(sample_np - x), axis=1)
    return int(np.argmin(a))


def nearest_debug_indices(x):
    a = np.sum(np.square(sample_np - x), axis=1)
    return a.argsort()


neighbourIndices = tree.query(test)

neighbourIndices_np = [nearest_debug(x) for x in test]

assert neighbourIndices == neighbourIndices_np, "Errors found in query"
for x in test:
    assert np.any(
        nearest_debug_indices(x)[:10] == tree.queryK(x, 10, True)
    ), "Errors found in queryK"

if ot.PlatformInfo.HasFeature("nanoflann"):
    radius = 0.3
    indices, distance = tree.queryRadius([0.0] * 3, radius, True)
    print(indices, distance)
    assert len(indices) == len(distance)
    for i, x in enumerate(sample):
        assert (i in indices) == (x.norm() < radius), f"{x.norm()}"
        if i in indices:
            ott.assert_almost_equal(x.norm(), distance[indices.index(i)])

# limit case
sample = [
    [0.0, -1.0 / 3.0],
    [-0.5, -1.0 / 3.0],
    [0.5, -0.77777777777],
    [-0.75, -0.111111111],
    [0.25, 0.55555555555],
    [-0.25, -0.55555555555],
]
tree = ot.KDTree(sample)
x = sample[1]
indices = tree.queryK(x, 6, True)
print(indices)
assert indices == [1, 5, 3, 0, 2, 4]



"""
Optimization of the Rastrigin test function
===========================================
"""

# %%
# The Rastrigin function is defined by:
#
# .. math::
#
#   f(\vect{x}) = A + \sum_{i=1}^n \left[x_i^2 - A\cos(2 \pi x_i)\right]
#
# where :math:`A=10` and :math:`\vect{x}\in[-5.12,5.12]^n`.
#
# It has a global minimum at :math:`\vect{x} = \vect{0}` where :math:`f(\vect{x})= - 10`.
#
# This function has many local minima, so optimization algorithms must be run from multiple starting points.
#
# In our example, we consider the bidimensional case, i.e. :math:`n=2`.
#
# **References**:
#
# - Rastrigin, L. A. "Systems of extremal control." Mir, Moscow (1974).
# - Rudolph, G. "Globale Optimierung mit parallelen Evolutionsstrategien". Diplomarbeit. Department of Computer Science, University of Dortmund, July 1990.
#

# %%
# Definition of the problem
# -------------------------

# %%
import openturns as ot
import openturns.viewer as otv
import math as m


def rastriginPy(X):
    A = 10.0
    delta = [x**2 - A * m.cos(2 * m.pi * x) for x in X]
    y = A + sum(delta)
    return [y]


dim = 2
rastrigin = ot.PythonFunction(dim, 1, rastriginPy)
print(rastrigin([1.0, 1.0]))

# %%
# Making `rastrigin` into a :class:`~openturns.MemoizeFunction` will make it recall all evaluated points.

# %%
rastrigin = ot.MemoizeFunction(rastrigin)

# %%
# This example is academic and the point achieving the global minimum of the function is known.

# %%
xexact = [0.0] * dim
print(xexact)

# %%
# The optimization bounds must be specified.

# %%
lowerbound = [-4.4] * dim
upperbound = [5.12] * dim
bounds = ot.Interval(lowerbound, upperbound)

# %%
# Plot the iso-values of the objective function
# ---------------------------------------------

# %%
graph = rastrigin.draw(lowerbound, upperbound, [100] * dim)
graph.setTitle("Rastrigin function")
graph.setLegendPosition("upper left")
graph.setLegendCorner([1.0, 1.0])
view = otv.View(graph)

# %%
# We see that the Rastrigin function has several local minima. However, there is only one single global minimum at :math:`\vect{x}^\star=(0, 0)`.

# %%
# Create the problem and set the optimization algorithm
# -----------------------------------------------------

# %%
problem = ot.OptimizationProblem(rastrigin)

# %%
# We use the :class:`~openturns.Cobyla` algorithm and run it from multiple starting points selected by a :class:`~openturns.LowDiscrepancyExperiment`.

# %%
size = 64
distribution = ot.JointDistribution([ot.Uniform(lowerbound[0], upperbound[0])] * dim)
experiment = ot.LowDiscrepancyExperiment(ot.SobolSequence(), distribution, size)
solver = ot.MultiStart(ot.Cobyla(problem), experiment.generate())

# %%
# Visualize the starting points of the optimization algorithm
# -----------------------------------------------------------

# %%
startingPoints = solver.getStartingSample()
graph = rastrigin.draw(lowerbound, upperbound, [100] * dim)
graph.setTitle("Rastrigin function")
cloud = ot.Cloud(startingPoints)
cloud.setPointStyle("bullet")
cloud.setColor("black")
graph.add(cloud)
graph.setLegends([""])
# sphinx_gallery_thumbnail_number = 2
view = otv.View(graph)

# %%
# We see that the starting points are well spread across the input domain of the function.

# %%
# Solve the optimization problem
# ------------------------------

# %%
solver.run()
result = solver.getResult()
xoptim = result.getOptimalPoint()
print(xoptim)

# %%
xexact

# %%
# We can see that the solver found a very accurate approximation of the exact solution.

# %%
# Analyze the optimization process
# --------------------------------
#
# :class:`~openturns.MultiStart` ran an instance of :class:`~openturns.Cobyla` from each starting point.
#
# Let us focus on the instance that found the global minimum. How many times did it evaluate `rastrigin`?

# %%
result.getCallsNumber()

# %%
# Let us view these evaluation points.

# %%
inputSample = result.getInputSample()
graph = rastrigin.draw(lowerbound, upperbound, [100] * dim)
graph.setTitle("Rastrigin function")
cloud = ot.Cloud(inputSample)
cloud.setPointStyle("bullet")
cloud.setColor("black")
graph.add(cloud)
graph.setLegendCorner([1.0, 1.0])
graph.setLegendPosition("upper left")
view = otv.View(graph)

# %%
# How fast did it find the global minimum?

# %%
graph = result.drawOptimalValueHistory()
view = otv.View(graph)

# %%
# Let us now analyze the :class:`~openturns.MultiStart` process as a whole.
#
# Since `rastrigin` is a :class:`~openturns.MemoizeFunction`,
# it has a :meth:`~openturns.MemoizeFunction.getInputHistory` method
# which lets us see all points it was evaluated on since its creation.

# %%
inputSample = rastrigin.getInputHistory()
graph = rastrigin.draw(lowerbound, upperbound, [100] * dim)
graph.setTitle("Rastrigin function")
cloud = ot.Cloud(inputSample)
cloud.setPointStyle("bullet")
cloud.setColor("black")
graph.add(cloud)
graph.setLegendCorner([1.0, 1.0])
graph.setLegendPosition("upper left")
view = otv.View(graph)

# %%
# How many times did all :class:`~openturns.Cobyla` instances combined call `rastrigin`?
print(rastrigin.getInputHistory().getSize())

# %%
otv.View.ShowAll()



#! /usr/bin/env python

import openturns as ot
import openturns.testing as ott

ot.TESTPREAMBLE()


# Construction based on two functions
# Create an instance */
myFunc = ot.PointToPointConnection(
    ot.SymbolicFunction("x", "sin(x)"), ot.SymbolicFunction("x", "cos(x)")
)

print("myFunc=", myFunc)
# Get the input and output description */
print("myFunc input description=", myFunc.getInputDescription())
print("myFunc output description=", myFunc.getOutputDescription())
# Get the input and output dimension */
print("myFunc input dimension=", myFunc.getInputDimension())
print("myFunc output dimension=", myFunc.getOutputDimension())
# Connection on a point */
point = [1.0] * myFunc.getInputDimension()
print("point=", point)
print("myFunc(point)=", myFunc(point))
print("myFunc.gradient(point)=", myFunc.gradient(point))
print("myFunc.hessian(point)=", myFunc.hessian(point))
# Get the number of calls */
print("called ", myFunc.getEvaluationCallsNumber(), " times")

# Construction based on a PointToFieldFunction followed by a FieldToPointFunction
# Create a KarhunenLoeveResult
mesh = ot.IntervalMesher([9]).build(ot.Interval(-1.0, 1.0))
cov1D = ot.AbsoluteExponential([1.0])
algo = ot.KarhunenLoeveP1Algorithm(mesh, cov1D, 1.0e-2)
algo.run()
result = algo.getResult()
# Create a PointToFieldFunction
lifting = ot.KarhunenLoeveLifting(result)
# Create a FieldToPointFunction
projection = ot.KarhunenLoeveProjection(result)
# Create an instance */
myFunc = ot.PointToPointConnection(projection, lifting)

print("myFunc=", myFunc)
# Get the input and output description */
print("myFunc input description=", myFunc.getInputDescription())
print("myFunc output description=", myFunc.getOutputDescription())
# Get the input and output dimension */
print("myFunc input dimension=", myFunc.getInputDimension())
print("myFunc output dimension=", myFunc.getOutputDimension())
# Connection on a point */
point = [1.0] * myFunc.getInputDimension()
print("point=", point)
print("myFunc(point)=", myFunc(point))
ott.assert_almost_equal(myFunc.gradient(point), ot.IdentityMatrix(8))
ott.assert_almost_equal(myFunc.hessian(point), ot.SymmetricTensor(8, 8), 0.0, 1e-7)
# Get the number of calls */
print("called ", myFunc.getEvaluationCallsNumber(), " times")



#! /usr/bin/env python

"""
This validation purpose is to check that the operator() yields a kronecker product
between a fixed covariance and embedded kernel (rho) evaluation
We make comparison point by point.
For discretize operator, we check hin unit tests that everything is ok
"""

import openturns as ot
import openturns.testing as ott

ot.TESTPREAMBLE()
ot.RandomGenerator.SetSeed(0)


def convert_hmat_to_matrix(hmat):
    res = ot.Matrix(hmat.getNbRows(), hmat.getNbColumns())
    for i in range(hmat.getNbRows()):
        x = ot.Point(hmat.getNbColumns())
        x[i] = 1.0
        y = ot.Point(hmat.getNbRows())
        hmat.gemv("N", 1.0, x, 0.0, y)
        for j in range(hmat.getNbColumns()):
            res[i, j] = y[j]
    return res


if __name__ == "__main__":

    models = []
    refModels = []
    # rho correlation
    scale = [4, 5]
    rho = ot.GeneralizedExponential(scale, 1)

    # Amplitude values
    amplitude = [1, 2]
    myModel = ot.KroneckerCovarianceModel(rho, amplitude)
    models.append(myModel)

    spatialCorrelation = ot.CorrelationMatrix(2)
    spatialCorrelation[0, 1] = 0.8
    myModel = ot.KroneckerCovarianceModel(rho, amplitude, spatialCorrelation)
    models.append(myModel)

    spatialCovariance = ot.CovarianceMatrix(2)
    spatialCovariance[0, 0] = 4.0
    spatialCovariance[1, 1] = 5.0
    spatialCovariance[1, 0] = 1.2

    myModel = ot.KroneckerCovarianceModel(rho, spatialCovariance)
    models.append(myModel)

    mesher = ot.IntervalMesher([5, 5])
    lowerbound = [-1, -1]
    upperBound = [2.0, 4.0]
    interval = ot.Interval(lowerbound, upperBound)
    mesh = mesher.build(interval)
    vertices = mesh.getVertices()

    for k, myModel in enumerate(models):
        sigma = myModel.getAmplitude()
        correlation = myModel.getOutputCorrelation()
        covariance = ot.CovarianceMatrix(correlation.getDimension())
        for j in range(correlation.getDimension()):
            covariance[j, j] = sigma[j] * sigma[j]
            for i in range(j + 1, correlation.getDimension()):
                covariance[i, j] = sigma[i] * sigma[j] * correlation[i, j]
        covariance.checkSymmetry()
        # Assert that models are the same
        for vertex in vertices:
            value = rho.computeAsScalar(vertex)
            ott.assert_almost_equal(myModel(vertex), covariance * value, 1e-16, 1e-15)

        C = myModel.discretize(vertices)
        C.checkSymmetry()
        # HMat checks
        p = ot.HMatrixParameters()
        p.setAssemblyEpsilon(1e-16)
        p.setRecompressionEpsilon(1e-16)
        hmat = myModel.discretizeHMatrix(vertices, p)
        M = convert_hmat_to_matrix(hmat)
        ott.assert_almost_equal(C, M, 1e-16, 1e-15)



#! /usr/bin/env python

import openturns as ot
import openturns.testing as ott

ot.TESTPREAMBLE()

# Test default constructor
print("experiment0=", repr(ot.GaussProductExperiment().generate()))

distribution = ot.JointDistribution([ot.Exponential(), ot.Triangular(-1.0, -0.5, 1.0)])
marginalSizes = [3, 6]
# Test the constructor based on marginal degrees
print("experiment1=", ot.GaussProductExperiment(marginalSizes))
# Test the constructor based on distribution
print("experiment2=", ot.GaussProductExperiment(distribution))
# Test the constructor based on marginal degrees and distribution
experiment = ot.GaussProductExperiment(distribution, marginalSizes)
print("experiment = ", experiment)
sample, weights = experiment.generateWithWeights()
print("sample = ", repr(sample))
print("weights = ", repr(weights))


# Check polynomial degree of exactness
def checkPolynomialExactness(
    marginalDegrees,
    marginalSizes,
    lowerBound=0.0,
    upperBound=1.0,
    rtol=1.0e-15,
    atol=0.0,
    verbose=False,
):
    """
    Check polynomial exactness of Gauss tensor product quadrature

    Parameters
    ----------
    marginalDegrees : list of int
        The polynomial degree of the marginal polynomials to integrate
    marginalSizes : list of int
        The number of nodes on each marginal axis.
    lowerBound : float
        The lower bound of quadrature
    upperBound : float
        The upper bound of quadrature
    rtol : float, > 0
        The relative tolerance
    atol : float, > 0
        The absolute tolerance
    verbose : bool
        Set to True to print intermediate messages

    Examples
    --------
    marginalDegrees = [5, 3, 7]
    marginalSizes = [3, 2, 4]  # Polynomial exactness space = P5 x P3 x P7
    checkPolynomialExactness(marginalDegrees, marginalSizes)
    """
    dimension = len(marginalDegrees)
    if len(marginalSizes) != dimension:
        raise ValueError(
            f"Number of marginal degrees is {dimension} "
            f"but number of marginal sizes is {len(marginalSizes)}"
        )

    # Set bounds
    bounds = ot.Interval([lowerBound] * dimension, [upperBound] * dimension)

    # Create polynomial
    polynomialCollection = ot.PolynomialCollection()
    for i in range(dimension):
        coefficients = [0.0] * (1 + marginalDegrees[i])
        coefficients[-1] = 1
        polynomial = ot.UniVariatePolynomial(coefficients)
        polynomialCollection.add(polynomial)

    productPoly = ot.ProductPolynomialEvaluation(polynomialCollection)

    # Create Gauss tensor product quadrature
    lowerBoundPoint = bounds.getLowerBound()
    upperBoundPoint = bounds.getUpperBound()
    distribution = ot.JointDistribution(
        [ot.Uniform(lowerBoundPoint[i], upperBoundPoint[i]) for i in range(dimension)]
    )
    experiment = ot.GaussProductExperiment(distribution, marginalSizes)

    # Evaluate integral
    nodes, weights = experiment.generateWithWeights()
    values = productPoly(nodes).asPoint()
    computedIntegral = weights.dot(values)

    # Expected integral
    expectedIntegral = 1.0
    for i in range(dimension):
        marginalIntegral = (
            upperBoundPoint[i] ** (1 + marginalDegrees[i])
            - lowerBoundPoint[i] ** (1 + marginalDegrees[i])
        ) / (1 + marginalDegrees[i])
        expectedIntegral *= marginalIntegral
    absoluteError = abs(computedIntegral - expectedIntegral)
    if verbose:
        print(
            f"Polynomial : {str(productPoly):20s}, "
            f"  integral computed  = {computedIntegral:.7f}, "
            f"  expected = {expectedIntegral:.7f}, "
            f"  absolute error = {absoluteError:.3e}"
        )
    ott.assert_almost_equal(expectedIntegral, computedIntegral, rtol, atol)
    return


# Test different polynomials, up to the maximum
marginalSizes = [3, 2, 4]  # Polynomial exactness space = P5 x P3 x P7
maximumMarginalDegrees = [2 * n for n in marginalSizes]
experiment = ot.Tuples(maximumMarginalDegrees)
marginalDegreesList = experiment.generate()
for i in range(marginalDegreesList.getSize()):
    marginalDegrees = marginalDegreesList[i]
    checkPolynomialExactness(
        marginalDegrees, marginalSizes, rtol=1.0e-14, verbose=False
    )



#! /usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()

# Default dimension parameter to evaluate the model
defaultDimension = 1
inputDimension = 1

# Amplitude values
amplitude = ot.Point(defaultDimension, 2.0)

# Scale values
scale = ot.Point(inputDimension, 1.0)

# Default constructor
myDefaultModel = ot.ExponentialModel()
print("myDefaultModel = ", myDefaultModel)

# Second order model with parameters
myModel = ot.ExponentialModel(scale, amplitude)
print("myModel = ", myModel)

timeValueOne = 1.0
print("covariance matrix at t = ", timeValueOne, " : ", myModel(timeValueOne))
print(
    "covariance matrix at t = ",
    -1.0 * timeValueOne,
    " : ",
    myModel(-1.0 * timeValueOne),
)

# Evaluation at time higher to check the decrease of the exponential values
timeValueHigh = 15.0
print(
    "covariance matrix at t = ", timeValueHigh, " : ", myModel(timeValueHigh).__str__()
)

timeGrid = ot.RegularGrid(0.0, 1.0 / 3.0, 4)
print(
    "discretized covariance over the time grid=",
    timeGrid,
    "is=",
    myModel.discretize(timeGrid),
)

# Default dimension parameter to evaluate the model
highDimension = 3

# Reallocation of adequate sizes
amplitude.resize(highDimension)
spatialCorrelation = ot.CorrelationMatrix(highDimension)
for index in range(highDimension):
    amplitude[index] = (index + 1) / (defaultDimension * defaultDimension)
    if index > 0:
        spatialCorrelation[index, index - 1] = 1.0 / (index * index)

# Second order model  - dimension 10
myHighModel = ot.ExponentialModel(scale, amplitude, spatialCorrelation)
print("myHighModel = ", myHighModel)

print("covariance matrix at t = ", timeValueOne, " : ", myHighModel(timeValueOne))
print(
    "covariance matrix at t = ",
    -1.0 * timeValueOne,
    " : ",
    myHighModel(-1.0 * timeValueOne),
)
print("covariance matrix at t = ", timeValueHigh, " : ", myHighModel(timeValueHigh))

print(
    "discretized covariance over the time grid=",
    timeGrid,
    "is=",
    myHighModel.discretize(timeGrid),
)

marginal = myHighModel.getMarginal([0, 2])
print("parameters=", myHighModel.getParameter(), myHighModel.getParameterDescription())
print(
    "marginal=",
    marginal,
    "marginal.parameter=",
    marginal.getParameter(),
    marginal.getParameterDescription(),
)

# parameter bug
model = ot.ExponentialModel([1.0] * 3, [2.0] * 2)
model.setActiveParameter(range(6))
p = model.getParameter()
print(p)
p[-1] = 0.5
model.setParameter(p)
p = model.getParameter()
print(p)

# draw bug
graph = ot.ExponentialModel().draw()
print("ok")



#! /usr/bin/env python

import openturns as ot


dim = 2

# We create an empty series
ts1 = ot.TimeSeries(0, dim)
ts1.setName("Ts1")

# We populate the empty ts
for p in range(3):
    pt = ot.Point(dim)
    for i in range(dim):
        pt[i] = 10.0 * (p + 1) + i
    ts1.add(pt)

print("ts1=", ts1)
print("len=", len(ts1))

# We get the second element of the ts
secondElement = ts1[1]
print("second element=", secondElement)

# We set the third element to a valid new element
newPoint = ot.Point(dim + 1)
for i in range(dim):
    newPoint[i + 1] = 1000.0 * (i + 1)
ts1[2] = newPoint
print("ts1=", ts1)

# We get the values of the second element of the ts
values1 = ts1.getValueAtIndex(1)
print("values of the second element=", values1)

# We set new values for the second element of the ts
newValues = values1 * 2.0
ts1.setValueAtIndex(1, newValues)
values2 = ts1.getValueAtIndex(1)
print("ts1=", ts1)
print("modified values of the second element=", values2, " should be ", newValues)

# We get the values of the second element of the ts
tree = ot.KDTree(ts1.getMesh().getVertices())
values3 = ts1.getValueAtIndex(tree.query([-1.0]))
print("values at t=-1.0 =", values3)
values4 = ts1.getValueAtIndex(tree.query([1.45]))
print("values at t=1.45 =", values4)
values5 = ts1.getValueAtIndex(tree.query([1.54]))
print("values at t=1.54 =", values5)
values6 = ts1.getValueAtIndex(tree.query([14.5]))
print("values at t=14.5 =", values6)

# We set new values for the third element of the ts
ts1.setValueAtIndex(tree.query([1.54]), values6 * -1.0)
print("ts1=", ts1)

ts2 = ot.TimeSeries(10, dim)
print("ts2=", ts2)

# ts2[5] = point2
# print 'ts2=', ts2

try:
    # We get the tenth element of the ts
    # THIS SHOULD NORMALLY FAIL
    tenthElement = ts1.at(9)
except Exception:
    print("Expected failure")

tg1 = ot.RegularGrid(0.0, 0.1, 11)
ts3 = ot.TimeSeries(tg1, dim)
print("ts3=", ts3)

tg2 = ot.RegularGrid(0.0, 0.2, 6)
ts4 = ot.TimeSeries(tg2, dim)
print("ts4=", ts4)

# We append a sample to a time series
ts5 = ot.TimeSeries(3, dim)
ns1 = ot.Sample(3, [99.9] * dim)
print("ts5=", ts5)
ts5.add(ns1)
print("ts5=", ts5)

# We retrieve the values of the time series as a sample
ns2 = ts5.getValues()
print("ns2=", ns2)

# We build a time series from a time grid and a sample
ns3 = ts4.getValues()
ts6 = ot.TimeSeries(tg2, ns3)
print("ts6=", ts6)



#! /usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()

# TEST NUMBER ZERO : DEFAULT CONSTRUCTOR AND STRING CONVERTER
print("test number zero : default constructor and string converter")

# Default constructor
tensor0 = ot.ComplexTensor()

# String converter
print("tensor0 = ", repr(tensor0))

# TEST NUMBER ONE : CONSTRUCTOR WITH SIZE, OPERATOR() AND STRING CONVERTER
print("test number one : constructor with size, operator() and string converter")

# Constructor with size
tensor1 = ot.ComplexTensor(2, 2, 3)

# Check operator() methods
tensor1[0, 0, 0] = 1 + 0.1j
tensor1[1, 0, 0] = 2 + 0.2j
tensor1[0, 1, 0] = 3 + 0.3j
tensor1[1, 1, 0] = 4 + 0.4j
tensor1[0, 0, 1] = 5 + 0.5j
tensor1[1, 0, 1] = 6 + 0.6j
tensor1[0, 1, 1] = 7 + 0.7j
tensor1[1, 1, 1] = 8 + 0.8j
tensor1[0, 0, 2] = 9 + 0.9j
tensor1[1, 0, 2] = 10 + 1.0j
tensor1[0, 1, 2] = 11 + 1.1j
tensor1[1, 1, 2] = 12 + 1.2j

# String converter
print("tensor1 = ", repr(tensor1))
print(
    "values = (%.6f, %.6fj)" % (tensor1[0, 0, 0].real, tensor1[0, 0, 0].imag),
    "  (%.6f, %.6fj)" % (tensor1[1, 0, 0].real, tensor1[1, 0, 0].imag),
    "  (%.6f, %.6fj)" % (tensor1[0, 1, 0].real, tensor1[0, 1, 0].imag),
    "  (%.6f, %.6fj)" % (tensor1[1, 1, 0].real, tensor1[1, 1, 0].imag),
    "  (%.6f, %.6fj)" % (tensor1[0, 0, 1].real, tensor1[0, 0, 1].imag),
    "  (%.6f, %.6fj)" % (tensor1[1, 0, 1].real, tensor1[1, 0, 1].imag),
    "  (%.6f, %.6fj)" % (tensor1[0, 1, 1].real, tensor1[0, 1, 1].imag),
    "  (%.6f, %.6fj)" % (tensor1[1, 1, 1].real, tensor1[1, 1, 1].imag),
    "  (%.6f, %.6fj)" % (tensor1[0, 0, 2].real, tensor1[0, 0, 2].imag),
    "  (%.6f, %.6fj)" % (tensor1[1, 0, 2].real, tensor1[1, 0, 2].imag),
    "  (%.6f, %.6fj)" % (tensor1[0, 1, 2].real, tensor1[0, 1, 2].imag),
    "  (%.6f, %.6fj)" % (tensor1[1, 1, 2].real, tensor1[1, 1, 2].imag),
)

# TEST NUMBER TWO : COPY CONSTRUCTOR AND STRING CONVERTER
print("test number two : copy constructor and string converter")

# Copy constructor
tensor2 = tensor1

# String converter
print("tensor2 = ", repr(tensor2))

# TEST NUMBER THREE : GET DIMENSIONS METHODS
print("test number three : get dimensions methods")

# Get dimension methods
print("tensor1's nbRows = ", tensor1.getNbRows(), end=" ")
print("tensor1's nbColumns = ", tensor1.getNbColumns())
print("tensor1's nbSheets = ", tensor1.getNbSheets())

# TEST NUMBER FOUR : ASSIGNMENT METHOD
print("test number four : assignment method")

# TEST NUMBER FIVE : ISEMPTY METHOD
print("test number five : isEmpty method")

# Check method isEmpty
tensor5 = ot.ComplexTensor()
print("tensor1 is empty = ", tensor1.isEmpty())
print("tensor0 is empty = ", tensor0.isEmpty())
print("tensor5 is empty = ", tensor5.isEmpty())

# TEST NUMBER SIX : GETSHEET AND SETSHEET METHODS
print("tensor1 = ", repr(tensor1))

#  ComplexMatrix sheet1(tensor1.getSheet(1))
print("tensor1.getSheet(1) = ", repr(tensor1.getSheet(1)))
sheet2 = ot.ComplexMatrix(2, 2)
sheet2[0, 0] = 0.5 - 0.5j
sheet2[1, 0] = 0.6 - 0.5j
sheet2[0, 1] = 0.7 - 0.5j
sheet2[1, 1] = 0.8 - 0.5j
print("sheet2 = ", repr(sheet2))
tensor1.setSheet(1, sheet2)
print("tensor1 = ", repr(tensor1))



"""
EfficientGlobalOptimization examples
====================================
"""

# %%
# The EGO algorithm [jones1998]_ is an adaptative optimization method based on
# Gaussian Process metamodel.
#
# An initial design of experiment is used to build a first metamodel.
# At each iteration a new point that maximizes a criterion is chosen as
# optimizer candidate.
#
# The criterion uses a tradeoff between the metamodel value and the conditional
# variance.
#
# Then the new point is evaluated using the original model and the metamodel is
# relearnt on the extended design of experiments.


# %%
from openturns.usecases import branin_function
from openturns.usecases import ackley_function
import openturns as ot
import openturns.experimental as otexp
import openturns.viewer as otv


# %%
# Ackley test-case
# ----------------
#
# We first apply the EGO algorithm on the :ref:`Ackley function<use-case-ackley>`.

# %%
# Define the problem
# ^^^^^^^^^^^^^^^^^^

# %%
# The Ackley model is defined in the usecases module in a data class `AckleyModel` :
am = ackley_function.AckleyModel()

# %%
# We get the Ackley function :
model = am.model

# %%
# We specify the domain of the model :
dim = am.dim
lowerbound = [-4.0] * dim
upperbound = [4.0] * dim

# %%
# We know that the global minimum is at the center of the domain. It is stored in the `AckleyModel` data class.
xexact = am.x0

# %%
# The minimum value attained `fexact` is :
fexact = model(xexact)
fexact

# %%
graph = model.draw(lowerbound, upperbound, [100] * dim)
graph.setTitle("Ackley function")
view = otv.View(graph)

# %%
# We see that the Ackley function has many local minimas. The global minimum, however, is unique and located at the center of the domain.

# %%
# Create the initial GP
# ^^^^^^^^^^^^^^^^^^^^^
#
# Before using the EGO algorithm, we must create an initial GP metamodel.
# In order to do this, we must create a design of experiment which fills the space.
# In this situation, the :class:`~openturns.LHSExperiment` is a good place to start (but other design of experiments may allow one to better fill the space).
# We use a uniform distribution in order to create a LHS design.
# The length of the first LHS is set to ten times the problem dimension as recommended in [jones1998]_.

# %%
listUniformDistributions = [
    ot.Uniform(lowerbound[i], upperbound[i]) for i in range(dim)
]
distribution = ot.JointDistribution(listUniformDistributions)
sampleSize = 10 * dim
experiment = ot.LHSExperiment(distribution, sampleSize)
inputSample = experiment.generate()
outputSample = model(inputSample)

# %%
graph = ot.Graph(
    "Initial LHS design of experiment - n=%d" % (sampleSize), "$x_0$", "$x_1$", True
)
cloud = ot.Cloud(inputSample)
graph.add(cloud)
view = otv.View(graph)

# %%
# We now create the GP metamodel.
# We selected the :class:`~openturns.MaternModel` covariance model with a constant basis as recommended in [leriche2021]_.

# %%
covarianceModel = ot.MaternModel([1.0] * dim, [0.5], 2.5)
basis = ot.ConstantBasisFactory(dim).build()
fitter = ot.GaussianProcessFitter(inputSample, outputSample, covarianceModel, basis)
fitter.run()
gpr = ot.GaussianProcessRegression(fitter.getResult())
gpr.run()

# %%
# Create the optimization problem
# ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
#
# We finally create the :class:`~openturns.OptimizationProblem` and solve it with class:`~openturns.EfficientGlobalOptimization`.

# %%
problem = ot.OptimizationProblem()
problem.setObjective(model)
bounds = ot.Interval(lowerbound, upperbound)
problem.setBounds(bounds)

# %%
# In order to show the various options, we configure them all, even if most could be left to default settings in this case.
#
# The most important method is :class:`~openturns.experimental.EfficientGlobalOptimization.setMaximumCallsNumber` which limits the number of iterations that the algorithm can perform.
# In the Ackley example, we choose to perform 30 iterations of the algorithm.

# %%
algo = otexp.EfficientGlobalOptimization(problem, gpr.getResult())
algo.setMaximumCallsNumber(30)
algo.run()
result = algo.getResult()

# %%
result.getIterationNumber()

# %%
result.getOptimalPoint()

# %%
result.getOptimalValue()

# %%
fexact

# %%
# Compared to the minimum function value, we see that the EGO algorithm
# provides solution which is accurate.
# Indeed, the optimal point is in the neighbourhood of the exact solution,
# and this is quite an impressive success given the limited amount of function
# evaluations: only 20 evaluations for the initial DOE and 30 iterations of
# the EGO algorithm, for a total equal to 50 function evaluations.

# %%
graph = result.drawOptimalValueHistory()
optimum_curve = ot.Curve(ot.Sample([[0, fexact[0]], [29, fexact[0]]]))
graph.add(optimum_curve)
view = otv.View(graph)

# %%
inputHistory = result.getInputSample()

# %%
graph = model.draw(lowerbound, upperbound, [100] * dim)
graph.setTitle("Ackley function")
cloud = ot.Cloud(inputSample, "initial")
cloud.setPointStyle("bullet")
cloud.setColor("black")
graph.add(cloud)
cloud = ot.Cloud(inputHistory, "solution")
cloud.setPointStyle("diamond")
cloud.setColor("forestgreen")
graph.add(cloud)
view = otv.View(graph)

# %%
# We see that the initial (black) points are dispersed in the whole domain,
# while the solution points are much closer to the solution.

# %%
# Branin test-case
# ----------------
#
# We now take a look at the :ref:`Branin-Hoo<use-case-branin>` function.

# %%
# Define the problem
# ^^^^^^^^^^^^^^^^^^

# %%
# The Branin model is defined in the usecases module in a data class `BraninModel` :
bm = branin_function.BraninModel()

# %%
# We load the dimension,
dim = bm.dim

# %%
# the domain boundaries,
lowerbound = bm.lowerbound
upperbound = bm.upperbound

# %%
# and we load the model function and its noise:
objectiveFunction = bm.model
noise = bm.trueNoiseFunction

# %%
# We build a sample out of the three minima :
xexact = ot.Sample([bm.xexact1, bm.xexact2, bm.xexact3])

# %%
# The minimum value attained `fexact` is :
fexact = objectiveFunction(xexact)
fexact

# %%
graph = objectiveFunction.draw(lowerbound, upperbound, [100] * dim)
graph.setTitle("Branin function")
view = otv.View(graph)

# %%
# The Branin function has three local minimas.

# %%
# Create the initial GP
# ^^^^^^^^^^^^^^^^^^^^^

# %%
distribution = ot.JointDistribution([ot.Uniform(0.0, 1.0)] * dim)
sampleSize = 10 * dim
experiment = ot.LHSExperiment(distribution, sampleSize)
inputSample = experiment.generate()
outputSample = objectiveFunction(inputSample)

# %%
graph = ot.Graph(
    "Initial LHS design of experiment - n=%d" % (sampleSize), "$x_0$", "$x_1$", True
)
cloud = ot.Cloud(inputSample)
graph.add(cloud)
view = otv.View(graph)

# %%
# Configure the covariance model with the noise
covarianceModel = ot.MaternModel([1.0] * dim, [0.5], 2.5)
covarianceModel.setNuggetFactor(noise)

# %%
# Build the initial GP
basis = ot.ConstantBasisFactory(dim).build()
fitter = ot.GaussianProcessFitter(inputSample, outputSample, covarianceModel, basis)
fitter.run()
gpr = ot.GaussianProcessRegression(fitter.getResult())
gpr.run()

# %%
# Create and solve the problem
# ^^^^^^^^^^^^^^^^^^^^^^^^^^^^

# %%
# We define the problem :
problem = ot.OptimizationProblem()
problem.setObjective(objectiveFunction)
bounds = ot.Interval(lowerbound, upperbound)
problem.setBounds(bounds)

# %%
# We configure the algorithm
# the nugget factor set in the covariance model with enable the AEI formulation
algo = otexp.EfficientGlobalOptimization(problem, gpr.getResult())
algo.setMaximumCallsNumber(30)

# %%
# We run the algorithm and get the result:
algo.run()
result = algo.getResult()

# %%
result.getIterationNumber()

# %%
result.getOptimalPoint()

# %%
result.getOptimalValue()

# %%
fexact

# %%
inputHistory = result.getInputSample()

# %%
graph = objectiveFunction.draw(lowerbound, upperbound, [100] * dim)
graph.setTitle("Branin function")
cloud = ot.Cloud(inputSample, "initial")
cloud.setPointStyle("bullet")
cloud.setColor("black")
graph.add(cloud)
cloud = ot.Cloud(inputHistory, "solution")
cloud.setPointStyle("diamond")
cloud.setColor("forestgreen")
graph.add(cloud)
view = otv.View(graph)

# %%
# We see that the EGO algorithm reached the different optima locations.

# %%
graph = result.drawOptimalValueHistory()
optimum_curve = ot.Curve(ot.Sample([[0, fexact[0][0]], [29, fexact[0][0]]]))
graph.add(optimum_curve)
view = otv.View(graph, axes_kw={"xticks": range(0, result.getIterationNumber(), 5)})


# %%
otv.View.ShowAll()



#! /usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()

# Poisson related functions
# dPoisson
lambdaMin = 0.2
lambdaMax = 5.0
kMax = 10
n1 = 5
for i1 in range(n1):
    lambda_ = lambdaMin + (lambdaMax - lambdaMin) * i1 / (n1 - 1)
    for k in range(kMax):
        print(
            "dPoisson(%.3g" % lambda_,
            ", ",
            k,
            ")=%.4g" % ot.DistFunc.dPoisson(lambda_, k),
        )
# dPoisson
for i1 in range(n1):
    lambda_ = lambdaMin + (lambdaMax - lambdaMin) * i1 / (n1 - 1)
    for k in range(kMax):
        print(
            "logdPoisson(%.3g" % lambda_,
            ", ",
            k,
            ")=%.4g" % ot.DistFunc.logdPoisson(lambda_, k),
        )
# rPoisson
nR = 5
for i1 in range(n1):
    lambda_ = lambdaMin + (lambdaMax - lambdaMin) * i1 / (n1 - 1)
    for iR in range(nR):
        print("rPoisson(%.3g" % lambda_, ")=", ot.DistFunc.rPoisson(lambda_))

# qPoisson
n2 = 20
for tail in [False, True]:
    for i1 in range(n1):
        lambda_ = lambdaMin + (lambdaMax - lambdaMin) * i1 / (n1 - 1)
        for i2 in range(n2):
            p = i2 / n2
            q0 = ot.DistFunc.qPoisson(lambda_, p, tail)
            print(f"qPoisson(lambda={lambda_:.3g}, p={p:.3g}, tail={tail})={q0}")



import openturns as ot
from matplotlib import pyplot as plt
import openturns.viewer as otv

Id = ot.IdentityMatrix(2)
atoms = [
    ot.Normal([1.0, 2.0], [0.5, 0.8], Id),
    ot.Normal([1.0, -2.0], [0.9, 0.8], Id),
    ot.Normal([-1.0, 0.0], [0.5, 0.6], Id),
]
weights = [0.3, 0.3, 0.4]
mixture = ot.Mixture(atoms, weights)
data = mixture.getSample(1000)
classifier = ot.MixtureClassifier(mixture)
graph = mixture.drawPDF(data.getMin(), data.getMax())
graph.setLegendPosition("")
graph.setTitle("MixtureClassifier example")
classes = classifier.classify(data)
palette = ot.Drawable.BuildDefaultPalette(len(atoms))
symbols = ot.Drawable.GetValidPointStyles()
for i in range(classes.getSize()):
    index = classes[i]
    graph.add(
        ot.Cloud(
            [data[i]], palette[index % len(palette)], symbols[index % len(symbols)]
        )
    )

fig = plt.figure(figsize=(4, 4))
axis = fig.add_subplot(111)
axis.set_xlim(auto=True)
otv.View(graph, figure=fig, axes=[axis], add_legend=False)



"""
Use a randomized QMC algorithm
==============================
"""

# %%
# In this example we are going to estimate a failure probability on the :ref:`stressed beam <use-case-stressed-beam>`.

# %%
from openturns.usecases import stressed_beam
import openturns as ot


# %%
# We load the data class containing the probabilistic modeling of the beam.
sm = stressed_beam.AxialStressedBeam()

# %%
# We load the joint probability distribution of the input parameters :
distribution = sm.distribution

# %%
# We load the model as well,
model = sm.model

# %%
# We create the event whose probability we want to estimate.

# %%
vect = ot.RandomVector(distribution)
G = ot.CompositeRandomVector(model, vect)
event = ot.ThresholdEvent(G, ot.Less(), 0.0)

# %%
# Define the low discrepancy sequence.

# %%
sequence = ot.SobolSequence()

# %%
# Create a simulation algorithm.

# %%
experiment = ot.LowDiscrepancyExperiment(sequence, 1)
experiment.setRandomize(True)
algo = ot.ProbabilitySimulationAlgorithm(event, experiment)
algo.setMaximumCoefficientOfVariation(0.05)
algo.setMaximumOuterSampling(int(1e4))
algo.run()

# %%
# Retrieve results.

# %%
result = algo.getResult()
probability = result.getProbabilityEstimate()
print("Pf=", probability)



#! /usr/bin/env python

import openturns as ot
import openturns.testing as ott
import openturns.experimental as otexp

ot.TESTPREAMBLE()

# Check the parameter constructor
mean = [0.0] * 2
std = [1.0] * 2
R = ot.CorrelationMatrix([[1.0, 0.8], [0.8, 1.0]])
distribution = ot.Normal(mean, std, R)
# These functions are obtained by using the Gram-Schmidt algorithm to make orthonormal
# wrt the distribution the first 2D tensorized Hermite polynomials
f0 = ot.SymbolicFunction(["x0", "x1"], ["1.0"])
f1 = ot.SymbolicFunction(["x0", "x1"], ["x0"])
f2 = ot.SymbolicFunction(["x0", "x1"], ["-4.0 * x0 / 3.0 + 5.0 * x1 / 3.0"])
f3 = ot.SymbolicFunction(["x0", "x1"], ["-1.0 / sqrt(2.0) + x0^2 / sqrt(2)"])
f4 = ot.SymbolicFunction(
    ["x0", "x1"],
    [
        "-4.0 / 3.0 - 1.885618083165693 * (-1.0 / sqrt(2.0) + x0^2 / sqrt(2.0)) + 5.0 * x0 * x1 / 3.0"
    ],
)
f5 = ot.SymbolicFunction(
    ["x0", "x1"],
    [
        "2.5141574442076222 + 16.0 / 9.0 * (-1.0 / sqrt(2.0) + x0^2 / sqrt(2.0)) - 3.142696805266918 * x0 * x1 + 25.0 / 9.0 * (-1.0 / sqrt(2.0) + x1^2 / sqrt(2.0))"
    ],
)
initialBasis = [f0, f1, f2, f3, f4, f5]
factory = otexp.FiniteOrthogonalFunctionFactory(initialBasis, distribution)
x = [0.5] * 2
kMax = len(initialBasis)
distribution = factory.getMeasure()
initialBasis = factory.getFunctionsCollection()
factory.setFunctionsCollection(initialBasis)
print("Factory=", factory)
print("Factory=", repr(factory))
functions = list()
for k in range(kMax):
    fk = factory.build(k)
    functions.append(fk)
    value = fk(x)
    print(f"FiniteOrthogonalFunction_{k}({x})={value}")

# Check orthonormality
M = ot.SymmetricMatrix(kMax)
integrationAlgo = ot.GaussLegendre([48] * 2)
for m in range(kMax):
    for n in range(m + 1):

        def wrapper(x):
            return functions[m](x) * functions[n](x)[0] * distribution.computePDF(x)

        kernel = ot.PythonFunction(distribution.getDimension(), 1, wrapper)
        value = integrationAlgo.integrate(kernel, distribution.getRange())[0]
        if abs(value) >= 1.0e-6:
            M[m, n] = value
ott.assert_almost_equal(M, ot.IdentityMatrix(kMax))



"""
Create a copula
===============
"""

# %%
# In this example we are going to create a Normal copula from a correlation matrix.
#

# %%
import openturns as ot


# %%
# Create the distribution
R = ot.CorrelationMatrix(3)
R[0, 1] = 0.25
R[1, 2] = 0.25
copula = ot.NormalCopula(R)
print(copula)



"""
Use the FORM algorithm in case of several design points
=======================================================
"""

# %%
# Abstract
# --------
#
# In this example we showcase the :class:`~openturns.MultiFORM` class which can perform a FORM analysis with several
# design points.

# %%
import openturns as ot
import openturns.viewer as otv

# %%
# We consider a standard bivariate Gaussian random vector :math:`X = (X_1, X_2)`:
dim = 2
dist = ot.Normal(dim)

# %%
# We can draw the bidimensional PDF of the distribution `dist` over :math:`[-5,5] \times [-5,5]`:
ot.ResourceMap.SetAsUnsignedInteger("Contour-DefaultLevelsNumber", 8)
graphPDF = dist.drawPDF([-5, -5], [5, 5])
graphPDF.setTitle(r"2D-PDF of the input variables $(X_1, X_2)$")
graphPDF.setXTitle(r"$x_1$")
graphPDF.setYTitle(r"$x_2$")
graphPDF.setLegendPosition("lower right")
view = otv.View(graphPDF, contour_kw={"norm": "log"})

# %%
# We then define a model :math:`f` which maps a 2D-vector X = (X_1,X_2) to a
# scalar output `Y = f(X)`.
f = ot.SymbolicFunction(["x0", "x1"], ["5.0-x1-0.5*(x0-0.1)^2"])
graphModel = f.draw([-8.0, -8.0], [8.0, 8.0])
graphModel.setXTitle(r"$x_1$")
graphModel.setXTitle(r"$x_2$")
graphModel.setTitle(r"Isolines of the model : $Y = f(X)$")
view = otv.View(graphModel)


# %%
# We create random vectors for the input and output variables:
X = ot.RandomVector(dist)
Y = ot.CompositeRandomVector(f, X)


# %%
# The failure domain :math:`\mathcal{D}` is :math:`\mathcal{D} = \{ x=(x_1, x_2) \in \mathbb{R}^2 / f(x) < 0 \}`
failureEvent = ot.ThresholdEvent(Y, ot.Less(), 0.0)


# %%
# We shall represent the failure domain event using a :class:`~openturns.Contour` object.
nx, ny = 25, 25
xx = ot.Box([nx], ot.Interval([-8.0], [8.0])).generate()
yy = ot.Box([ny], ot.Interval([-8.0], [8.0])).generate()
inputData = ot.Box([nx, ny], ot.Interval([-8.0, -8.0], [8.0, 8.0])).generate()
outputData = f(inputData)
mycontour = ot.Contour(xx, yy, outputData)
mycontour.setLevels([0.0])
mycontour.setLabels(["0.0"])
mycontour.setColor("black")
mycontour.setLineStyle("dashed")
graphModel.add(mycontour)
view = otv.View(graphModel)

# %%
# In the physical space the failure domain boundary is the dashed black curve. We recall that one of
# the steps of the FORM method is to find the closest point of the failure domain boundary to the origin.
# Here we see that the symmetry of the domain implies that two points exist;
# one in the :math:`x_1 \geq 0` half-space and the other in the :math:`x_1 \leq 0` half-space.


# %%
# We build the :class:`~openturns.MultiFORM` algorithm in a similar fashion as the :class:`~openturns.FORM` algorithm.
# We choose an optimization solver, here the Cobyla solver, and a starting point, the mean of the distribution `dist`.
solver = ot.Cobyla()
solver.setStartingPoint(dist.getMean())
algo = ot.MultiFORM(solver, failureEvent)


# %%
# We are ready to run the algorithm and store the result in the :class:`~openturns.MultiFORM` class:
algo.run()
result = algo.getResult()

# %%
# We have access to the results with the getFORMResultCollection method which produces a collection of :class:`~openturns.FORMResult`:
coll = result.getFORMResultCollection()

# %%
# The length of this collection is the number of design points:
n_design_pts = len(coll)
print("Number of design points :", n_design_pts)


# %%
# We have access to the design points with the getPhysicalSpaceDesignPoint method
# for each element of the collection `coll`.
designPointPhysicalSpace1 = coll[0].getPhysicalSpaceDesignPoint()
designPointPhysicalSpace2 = coll[1].getPhysicalSpaceDesignPoint()
print(coll[0].getPhysicalSpaceDesignPoint())
print(coll[1].getPhysicalSpaceDesignPoint())

# %%
# We visualize them on the previous graph with red circle dots.
cloud = ot.Cloud([designPointPhysicalSpace1[0]], [designPointPhysicalSpace1[1]])
cloud.setColor("red")
cloud.setPointStyle("fcircle")
cloud.setLegend("design point no. 1")
graphModel.add(cloud)

cloud = ot.Cloud([designPointPhysicalSpace2[0]], [designPointPhysicalSpace2[1]])
cloud.setColor("red")
cloud.setPointStyle("fcircle")
cloud.setLegend("design point no. 2")
graphModel.add(cloud)
graphModel.setLegendPosition("")
view = otv.View(graphModel)

# %%
# We recall that the FORM approximate is based on the substitution of the failure domain by the half-space defined by the tangent at the design point.
# Here we can clearly see that this would miss half of the information. That is why both design points are needed.

# %%
# For each design point we have a probability associated to the approximation by the half-space:
pf1 = coll[0].getEventProbability()
pf2 = coll[1].getEventProbability()

# %%
# The probability of failure is the given by the :meth:`~openturns.MultiFORMResult.getEventProbability`
# which is the sum of the two previous probabilities `pf1` and `pf2`:
pf = result.getEventProbability()
print("Probability of failure : ", pf)
print("    wrt design point 1 : ", pf1)
print("    wrt design point 2 : ", pf2)

# %%
# Display the figures
otv.View.ShowAll()



"""
Use the Directional Sampling Algorithm
======================================
"""

# %%
# In this example we estimate a failure probability with the directional simulation algorithm provided by the :class:`~openturns.DirectionalSampling` class.

# %%
# Introduction
# ------------
#
# The directional simulation algorithm  operates in the standard space based on:
#
# 1. a *root strategy* to evaluate the nearest failure point along each direction and take the contribution of each direction to the failure event probability into account.
#    The available strategies are:
#    - `RiskyAndFast`
#    - `MediumSafe`
#    - `SafeAndSlow`
#
# 2. a *sampling strategy* to choose directions in the standard space. The available strategies are:
#    - `RandomDirection`
#    - `OrthogonalDirection`
#
# Let us consider the analytical example of the cantilever beam described :ref:`here <use-case-cantilever-beam>`.
#

# %%
from openturns.usecases import cantilever_beam
import openturns as ot
import openturns.viewer as otv


# %%
# We load the model from the usecases module :
cb = cantilever_beam.CantileverBeam()

# %%
# We load the joint probability distribution of the input parameters :
distribution = cb.distribution

# %%
# We load the model giving the displacement at the end of the beam :
model = cb.model

# %%
# We create the event whose probability we want to estimate.

# %%
vect = ot.RandomVector(distribution)
G = ot.CompositeRandomVector(model, vect)
event = ot.ThresholdEvent(G, ot.Greater(), 0.30)

# %%
# Root finding algorithm.

# %%
solver = ot.Brent()
rootStrategy = ot.MediumSafe(solver)

# %%
# Direction sampling algorithm.

# %%
samplingStrategy = ot.OrthogonalDirection()

# %%
# Create a simulation algorithm.

# %%
algo = ot.DirectionalSampling(event, rootStrategy, samplingStrategy)
algo.setMaximumCoefficientOfVariation(0.1)
algo.setMaximumOuterSampling(40000)
algo.setConvergenceStrategy(ot.Full())
algo.run()

# %%
# Retrieve results.

# %%
result = algo.getResult()
probability = result.getProbabilityEstimate()
print("Pf=", probability)

# %%
# We can observe the convergence history with the `drawProbabilityConvergence`
# method.
graph = algo.drawProbabilityConvergence()
graph.setLogScale(ot.GraphImplementation.LOGX)
view = otv.View(graph)

# %%
# Display all figures
otv.View.ShowAll()



"""
Create a mixture of distributions
=================================
"""

# %%
#
# Introduction
# ------------
#
# In this example, we show how to build a distribution which is a Mixture from a collection of
# distributions of the same dimension :math:`d \geq 1`. We denote by :math:`\inputRV`
# the random vector with such a distribution.
#
# Let :math:`(\cL_1, \dots, \cL_N)` a collection of distributions and :math:`(\alpha_1, \dots,
# \alpha_N)` a collection of positive weights such that :math:`\sum_{i=1}^N \alpha_i = 1`.
# Let :math:`\inputRV_i` be a random vector whose distribution is :math:`\cL_i`.
#
# In the case where all the :math:`\cL_i` have a probability density function :math:`\mu_i`,
# then the mixture has a probability density function :math:`\inputMeasure` defined by:
#
# .. math::
#    \inputMeasure(\vect{x}) =  \sum_{i=1}^N \alpha_i \mu_i(\vect{x})
#
# In the case where all the :math:`\cL_i` are discrete, then the mixture is discrete and its
# probability distribution function is defined by:
#
# .. math::
#    \Prob{\inputRV = \vect{x}} =  \sum_{i=1}^N \alpha_i \Prob{\vect{X}_i = \vect{x}}
#
# In the case where some of the :math:`\cL_i`  have a probability density function and
# the other ones are discrete, then the mixture is not discrete and does not have a
# probability density function. Its cumulative distribution function is defined by:
#
# .. math::
#    F_\inputRV(\vect{x}) =  \sum_{i=1}^N \alpha_i F_{\vect{X}_i}(\vect{x})
#
# We illustrate the following particular cases:
#
# - Case 1a: Mixture of continuous distributions,
# - Case 1b: Mixture of copulas,
# - Case 1c: Mixture of a Histogram and a Generalized Pareto Distribution,
# - Case 2: Mixture of discrete distributions,
# - Case 3: Mixture of discrete and continuous distributions.

# %%
import openturns as ot
import openturns.viewer as otv


# %%
# Case 1a: Mixture of continuous distributions
# --------------------------------------------
# In this case, we build the mixture of the following continuous distributions:
#
# - a :class:`~openturns.Triangular`,
# - a :class:`~openturns.Normal`,
# - a :class:`~openturns.Uniform`.
#
# The weigths are automatically normalized.

# %%
# We define the collection of distributions and the associated weights.
distributions = [
    ot.Triangular(1.0, 2.0, 4.0),
    ot.Normal(-1.0, 1.0),
    ot.Uniform(5.0, 6.0),
]
weights = [0.4, 1.0, 0.2]

# %%
# We create the mixture.
distribution = ot.Mixture(distributions, weights)
print(distribution)

# %%
# We can draw the probability density function.
graph = distribution.drawPDF()
graph.setTitle("Mixture of Triangular, Normal, Uniform")
graph.setXTitle("x")
graph.setLegendPosition("")
view = otv.View(graph)

# %%
# We can draw the cumulative distribution function.
graph = distribution.drawCDF()
graph.setTitle("Mixture of Triangular, Normal, Uniform")
graph.setXTitle("x")
graph.setLegendPosition("")
view = otv.View(graph)

# %%
# Case 1b: Mixture of copulas
# ---------------------------
# In this case, we build the mixture of the following copulas:
#
# - a :class:`~openturns.GumbelCopula`,
# - a :class:`~openturns.ClaytonCopula`.

# %%
# We define the collection of copulas and the associated weights.
copulas = [ot.GumbelCopula(4.5), ot.ClaytonCopula(2.3)]
weights = [0.2, 0.8]

# %%
# We create a mixture of copulas.
distribution = ot.Mixture(copulas, weights)
print(distribution)

# %%
# We can draw the probability density function.
graph = distribution.drawPDF()
graph.setTitle("Mixture of Gumbel copula, Clayton copula")
graph.setXTitle(r"$x_0$")
graph.setYTitle(r"$x_1$")
graph.setLegendPosition("")
view = otv.View(graph)

# %%
# We can draw the cumulative distribution function.
graph = distribution.drawCDF()
graph.setTitle("Mixture of Gumbel copula, Clayton copula")
graph.setXTitle(r"$x_0$")
graph.setYTitle(r"$x_1$")
view = otv.View(graph)


# %%
# Case 1c: Mixture of a Histogram and a Generalized Pareto Distribution
# ---------------------------------------------------------------------
# We want to create the scalar distribution of :math:`X` such that:
#
# .. math::
#    X|X \leq x_0 & \sim \mathcal{L}_1 \\
#    X|X \geq x_0 & \sim \mathcal{L}_2
#
# where:
#
# - :math:`\mathcal{L}_1` is a Histogram,
# - :math:`\mathcal{L}_2` is a Generalized Pareto distribution (GPD),
# - :math:`x_0` is a quantile of high level of :math:`X`.
#
# Let us define:
#
# .. math::
#    w = \Prob{X \leq x_0}
#
# We assume that we only have a sample from :math:`X`.

# %%
# In this example, we consider a Normal distribution with zero mean and unit variance.
# We generate a sample of size :math:`n`.
n = 5000
X_dist = ot.Normal()
sample_X = X_dist.getSample(n)

# %%
# We build the whole histogram from the sample.
hist_dist = ot.HistogramFactory().build(sample_X)
g_hist = hist_dist.drawPDF()
g_hist.setTitle(r"Empirical distribution of $X$")
g_hist.setXTitle("x")
g_hist.setYTitle("pdf")
g_hist.setLegends(["histogram"])
view = otv.View(g_hist)

# %%
# We estimate the extreme empirical quantile of level 0.95.
w = 0.95
x0 = hist_dist.computeQuantile(w)[0]

# %%
# We start by truncating the initial histogram on the interval :math:`]-\infty, x_0]`.
# We visualize it!
hist_trunc = ot.TruncatedDistribution(hist_dist, x0, ot.TruncatedDistribution.UPPER)
g_hist_trunc = hist_trunc.drawPDF()
g_hist_trunc.setTitle(r"Empirical distribution of $X|X \leq $" + "%.3g" % (x0))
g_hist_trunc.setXTitle("x")
g_hist_trunc.setYTitle("pdf")
g_hist_trunc.setLegends(["truncated histogram"])
view = otv.View(g_hist_trunc)

# %%
# Then we model :math:`X|X \geq x_0` by a Generalized Pareto distribution (GPD). We start by
# extracting from the sample all the values greater than :math:`x_0` to build the upper
# sample. We get about :math:`n(1-w)` points.
sample_X_upper = ot.Sample(0, 1)
for i in range(len(sample_X)):
    if sample_X[i, 0] > x0:
        sample_X_upper.add(sample_X[i])

print("Excess number = ", sample_X_upper.getSize())
print("n(1-w) = ", int(n * (1 - w)))

# %%
# Then we fit a GPD parameterized by :math:`(\sigma, \xi, x_0)`: the
# threshold is fixed to :math:`x_0`.
# We use the estimator that maximizes the likelihood.
# To solve the optimisation problem faster, we start by estimating the 3 parameters
# :math:`(\sigma, \xi, u)` from the upper sample. Then we fix the threshold to :math:`u = x_0`
# and we estimate the remaining parameters :math:`(\sigma, \xi)` using the previous values of
# :math:`(\sigma, \xi)` as a starting point to the optimization problem.
# We visualize the pdf of the GPD.
gpd_first = ot.GeneralizedParetoFactory().build(sample_X_upper)
mlFact = ot.MaximumLikelihoodFactory(gpd_first)

# we fix the threshold to :math:`x_0`.
mlFact.setKnownParameter([2], [x0])
gpd_estimated = mlFact.build(sample_X_upper)
print("estimated gpd = ", gpd_estimated)

g_gpd = gpd_estimated.drawPDF()
g_gpd.setTitle(r"Distribution of $X|X \geq $" + "%.3g" % (x0))
g_gpd.setXTitle("x")
g_gpd.setYTitle("pdf")
g_gpd.setLegends(["GPd"])
view = otv.View(g_gpd)

# %%
# Then we can create the mixture using the truncated Histogram distribution below :math:`x_0`
# and the GPD over :math:`x_0` weighted by :math:`w` and :math:`(1-w)`.
mixt_dist = ot.Mixture([hist_trunc, gpd_estimated], [w, 1 - w])
g_hist.add(mixt_dist.drawPDF())

ord_Max = max(hist_dist.getImplementation().getHeight())
line = ot.Curve([x0, x0], [0.0, ord_Max])
line.setColor("red")
line.setLineStyle("dashed")
g_hist.add(line)

draw_ref = X_dist.drawPDF().getDrawable(0)
draw_ref.setLineStyle("dashed")
g_hist.add(draw_ref)

g_hist.setLegends(["histogram", "mixture", "", "exact dist"])
g_hist.setTitle(r"Distribution of $X$: Mixture")

view = otv.View(g_hist)

# We draw here only the mixture distribution to make the comparison easier.
g_mixt = mixt_dist.drawPDF()
g_mixt.setTitle(r"Mixture distribution of $X$")
g_mixt.setXTitle("x")
g_mixt.setYTitle("pdf")
g_mixt.setLegendPosition("")
view = otv.View(g_mixt)


# %%
# Case 2: Mixture of discrete distributions
# -----------------------------------------
# In this case, we build the mixture of the following distributions:
#
# - a :class:`~openturns.Poisson`,
# - a :class:`~openturns.Geometric`.
#
# The weigths are automatically normalized.

# %%
# We define the collection of distributions and the associated weights.
distributions = [ot.Poisson(1.2), ot.Geometric(0.7)]
weights = [0.4, 1.0]

# %%
# We create the mixture.
distribution = ot.Mixture(distributions, weights)
print(distribution)

# %%
# We can draw the probability distribution function.
graph = distribution.drawPDF()
graph.setTitle("Mixture of Poisson, Geometric")
graph.setXTitle("x")
graph.setLegendPosition("")
view = otv.View(graph)

# %%
# We can draw the cumulative distribution function.
graph = distribution.drawCDF()
graph.setTitle("Mixture of Poisson, Geometric")
graph.setXTitle("x")
graph.setLegendPosition("")
view = otv.View(graph)


# %%
# Case Case 3: Mixture of discrete and continuous distributions
# -------------------------------------------------------------
# In this case, we build the mixture of the following distributions:
#
# - a :class:`~openturns.Normal`,
# - a :class:`~openturns.Poisson`.
#
# The resulting distribution is not continuous nor discrete. It not possible to
# draw the pdf...

# %%
# We define the collection of distributions and the associated weights.
distributions = [ot.Normal(), ot.Poisson(0.7)]
weights = [0.4, 1.0]

# %%
# We create the mixture.
distribution = ot.Mixture(distributions, weights)
print(distribution)

# %%
# We cannot draw the probability distribution function as it is not defined.
# But, we can draw the cumulative distribution function.
graph = distribution.drawCDF()
graph.setTitle("Mixture of Normal, Poisson")
graph.setXTitle("x")
graph.setLegendPosition("")
view = otv.View(graph)

# %%
# Show all the graphs.
otv.View.ShowAll()



import openturns as ot
from matplotlib import pyplot as plt
import openturns.viewer as otv


# Generate sample with the given plane
distribution = ot.JointDistribution([ot.Uniform(0, 1)] * 2)
size = 200
experiment = ot.LowDiscrepancyExperiment(ot.SobolSequence(), distribution, size)

sample = experiment.generate()

# Create an empty graph
graph = ot.Graph("Low Discrepancy experiment", "x1", "x2")

# Create the cloud
cloud = ot.Cloud(sample, "blue", "plus", "")

# Then, draw it
graph.add(cloud)

fig = plt.figure(figsize=(4, 4))
axis = fig.add_subplot(111)
axis.set_xlim(auto=True)
otv.View(graph, figure=fig, axes=[axis], add_legend=False)




import gdb
import re
import sys
if sys.version > '3':
    long = int

objfile = gdb.current_objfile() or gdb.objfiles()[0]
int_ptr  = gdb.lookup_type('int').pointer()

ot_pretty_printers = []

def AddToPrettyPrinter(rx):
    def class_wrapper(cls):
        ot_pretty_printers.append( (re.compile(rx), cls) )
        return cls
    return class_wrapper

@AddToPrettyPrinter(r'^OT::PersistentObject$')
class PersistentObjectPrinter(object):
    def __init__(self, val):
        self.val = val

    def to_string(self):
        name = self.val['p_name_']['ptr_']['ptr_']
        if name == gdb.Value(0):
            return "[unnamed]"
        else:
            return "[name=%s]" % (str(name.dereference()))

@AddToPrettyPrinter(r'^OT::Pointer<.*>$')
class PointerPrinter(object):
    def __init__(self, val):
        self.val = val

    def to_string(self):
        global int_ptr

        pointee = self.val['ptr_']['ptr_']
        if long(pointee) == long(0):
            return "NULL"
        else:
            atomicVal = self.val['ptr_']['count_']['p_impl_']['use_count_']['val_']
            # This atomicVal points either to an int or a tbb::atomic<int>
            # The implementation of tbb::atomic depends on tbb's version, thus it is easier
            # to dereference it.
            atomicVal = atomicVal.address.cast(int_ptr).dereference()
            return "([counter=%s] %s) 0x%x = %s" % (str(atomicVal), str(pointee.type), long(pointee), str(pointee.dereference()))

@AddToPrettyPrinter(r'^OT::TypedInterfaceObject<.*>$')
class TypedInterfaceObjectPrinter(object):
    def __init__(self, val):
        self.val = val
        self.template_name = val.type.template_argument(0)

    def to_string(self):
        return "{ p_implementation_ -> [ %s ] }" % (str(self.val['p_implementation_']))

@AddToPrettyPrinter(r'^OT::(PersistentCollection<.*>|.*PersistentCollection)$')
class PersistentCollectionPrinter(object):
    def __init__(self, val):
        self.val = val

    def display_hint(self):
        return 'array'

    def to_string(self):
        # Dereference PersistentObject
        name = self.val['p_name_']['ptr_']['ptr_']
        if long(name) == long(0):
            decoratedName = "[unnamed]"
        else:
            decoratedName = "[name=%s]" % (str(name.dereference()))
        # Dereference Collection
        return "%s %s" % (decoratedName, self.val['coll__'])

@AddToPrettyPrinter(r'^OT::(Collection<.*>|.*Collection)$')
class CollectionPrinter(object):
    def __init__(self, val):
        self.val = val

    def display_hint(self):
        return 'array'

    def to_string(self):
        return str(self.val['coll__'])


def lookup_function(val):
    type = val.type
    if type.code == gdb.TYPE_CODE_REF:
        type = type.target()
    type = type.unqualified().strip_typedefs()
    typename = type.tag
    if not typename or not typename.startswith('OT::'):
        return None
    for regex, pretty_printer in ot_pretty_printers:
        if regex.search(typename):
            return pretty_printer(val)
    return None

objfile.pretty_printers.append(lookup_function)



#! /usr/bin/env python

import openturns as ot
import openturns.testing
import os
import math as m

ot.TESTPREAMBLE()

try:
    fileName = "myStudySaveLoad.xml"

    # Create a Study Object by name
    myStudy = ot.Study(fileName)
    point = ot.Point(2, 1.0)
    myStudy.add("point", point)
    myStudy.save()
    myStudy2 = ot.Study(fileName)
    myStudy2.load()
    point2 = ot.Point()
    myStudy2.fillObject("point", point2)
    # cleanup
    os.remove(fileName)

    # Create a Study Object with compression
    myStudy = ot.Study()
    compressionLevel = 5
    myStudy.setStorageManager(ot.XMLStorageManager(fileName + ".gz", compressionLevel))
    point = ot.Point(2, 1.0)
    myStudy.add("point", point)
    myStudy.save()
    myStudy2 = ot.Study(fileName + ".gz")
    myStudy2.load()
    point2 = ot.Point()
    myStudy2.fillObject("point", point2)
    # cleanup
    os.remove(fileName + ".gz")

    # Create a Study Object with compression, direct way
    compressionLevel = 5
    myStudy = ot.Study(fileName, compressionLevel)
    point = ot.Point(2, 1.0)
    myStudy.add("point", point)
    myStudy.save()
    myStudy2 = ot.Study(fileName)
    myStudy2.load()
    point2 = ot.Point()
    myStudy2.fillObject("point", point2)
    # cleanup
    os.remove(fileName)

    # Create a Study Object
    myStudy = ot.Study()
    myStudy.setStorageManager(ot.XMLStorageManager(fileName))

    # Add a PersistentObject to the Study (here a Point)
    p1 = ot.Point(3, 0.0)
    p1.setName("Good")
    p1[0] = 10.0
    p1[1] = 11.0
    p1[2] = 12.0
    myStudy.add(p1)

    # Add another PersistentObject to the Study (here a Sample)
    s1 = ot.Sample(3, 2)
    s1.setName("mySample")
    p2 = ot.Point(2, 0.0)
    p2.setName("One")
    p2[0] = 100.0
    p2[1] = 200.0
    s1[0] = p2
    p3 = ot.Point(2, 0.0)
    p3.setName("Two")
    p3[0] = 101.0
    p3[1] = 201.0
    s1[1] = p3
    p4 = ot.Point(2, 0.0)
    p4.setName("Three")
    p4[0] = 102.0
    p4[1] = 202.0
    s1[2] = p4
    myStudy.add("mySample", s1)

    # Add a point with a description
    pDesc = ot.PointWithDescription(p1)
    desc = pDesc.getDescription()
    desc[0] = "x"
    desc[1] = "y"
    desc[2] = "z"
    pDesc.setDescription(desc)
    myStudy.add(pDesc)

    # Add a matrix
    matrix = ot.Matrix(2, 3)
    matrix[0, 0] = 0
    matrix[0, 1] = 1
    matrix[0, 2] = 2
    matrix[1, 0] = 3
    matrix[1, 1] = 4
    matrix[1, 2] = 5
    myStudy.add("m", matrix)

    # Create a Point that we will try to reinstaciate after reloading
    point = ot.Point(2, 1000.0)
    point.setName("point")
    myStudy.add("point", point)

    # Create a Simulation::Result
    simulationResult = ot.ProbabilitySimulationResult(
        ot.ThresholdEvent(), 0.5, 0.01, 150, 4
    )
    myStudy.add("simulationResult", simulationResult)

    cNameList = [
        "DirectionalSampling",
        "SimulationSensitivityAnalysis",
        "ProbabilitySimulationAlgorithm",
    ]
    for cName in cNameList:
        otClass = getattr(ot, cName)
        instance = otClass()
        print("--", cName, instance)
        myStudy.add(cName, instance)

    # Create a Beta distribution
    beta = ot.Beta(3.0, 2.0, -1.0, 4.0)
    myStudy.add("beta", beta)

    # Create an analytical Function
    input = ot.Description(3)
    input[0] = "a"
    input[1] = "b"
    input[2] = "c"
    formulas = ot.Description(3)
    formulas[0] = "a+b+c"
    formulas[1] = "a-b*c"
    formulas[2] = "(a+2*b^2+3*c^3)/6"
    analytical = ot.SymbolicFunction(input, formulas)
    analytical.setName("analytical")
    analytical.setOutputDescription(["z1", "z2", "z3"])
    myStudy.add("analytical", analytical)

    # Create a TaylorExpansionMoments algorithm
    antecedent = ot.RandomVector(ot.IndependentCopula(analytical.getInputDimension()))
    antecedent.setName("antecedent")
    composite = ot.CompositeRandomVector(analytical, antecedent)
    composite.setName("composite")
    taylorExpansionsMoments = ot.TaylorExpansionMoments(composite)
    taylorExpansionsMoments.setName("taylorExpansionsMoments")
    taylorExpansionsMoments.getMeanFirstOrder()
    taylorExpansionsMoments.getMeanSecondOrder()
    taylorExpansionsMoments.getCovariance()

    myStudy.add("taylorExpansionsMoments", taylorExpansionsMoments)

    # Create a FORMResult
    input2 = ot.Description(2)
    input2[0] = "x"
    input2[1] = "y"
    formula2 = ot.Description(1)
    formula2[0] = "y^2-x"
    model = ot.SymbolicFunction(input2, formula2)
    model.setName("sum")
    input3 = ot.RandomVector(ot.Normal(2))
    input3.setName("input")
    output3 = ot.CompositeRandomVector(model, input3)
    output3.setName("output")
    event = ot.ThresholdEvent(output3, ot.Greater(), 1.0)
    event.setName("failureEvent")
    designPoint = ot.Point(2, 0.0)
    designPoint[0] = 1.0
    formResult = ot.FORMResult(ot.Point(2, 1.0), event, False)
    formResult.setName("formResult")
    formResult.getImportanceFactors()
    formResult.getEventProbabilitySensitivity()
    myStudy.add("formResult", formResult)

    # Create a SORMResult
    sormResult = ot.SORMResult([1.0] * 2, event, False)
    sormResult.setName("sormResult")
    sormResult.getEventProbabilityBreitung()
    sormResult.getEventProbabilityHohenbichler()
    sormResult.getEventProbabilityTvedt()
    sormResult.getGeneralisedReliabilityIndexBreitung()
    sormResult.getGeneralisedReliabilityIndexHohenbichler()
    sormResult.getGeneralisedReliabilityIndexTvedt()
    myStudy.add("sormResult", sormResult)

    # Create a RandomGeneratorState
    ot.RandomGenerator.SetSeed(0)
    randomGeneratorState = ot.RandomGeneratorState(ot.RandomGenerator.GetState())
    myStudy.add("randomGeneratorState", randomGeneratorState)

    # Create a GeneralLinearModelResult
    generalizedLinearModelResult = ot.GeneralLinearModelResult()
    generalizedLinearModelResult.setName("generalizedLinearModelResult")
    myStudy.add("generalizedLinearModelResult", generalizedLinearModelResult)

    # KDTree
    sample = ot.Normal(3).getSample(10)
    kDTree = ot.KDTree(sample)
    myStudy.add("kDTree", kDTree)

    # Distribution parameters

    # ArcsineMuSigma parameter ave
    ams_parameters = ot.ArcsineMuSigma(8.4, 2.25)
    myStudy.add("ams_parameters", ams_parameters)
    # BetaMuSigma parameter save
    bms_parameters = ot.BetaMuSigma(0.2, 0.6, -1, 2)
    myStudy.add("bms_parameters", bms_parameters)
    # GammaMuSigma parameter save
    gmms_parameters = ot.GammaMuSigma(1.5, 2.5, -0.5)
    myStudy.add("gmms_parameters", gmms_parameters)
    # GumbelMuSigma parameter save
    gms_parameters = ot.GumbelMuSigma(1.5, 1.3)
    myStudy.add("gms_parameters", gms_parameters)
    # LogNormalMuSigma parameter save
    lnms_parameters = ot.LogNormalMuSigma(30000.0, 9000.0, 15000)
    myStudy.add("lnms_parameters", lnms_parameters)
    # LogNormalMuSigmaOverMu parameter save
    lnmsm_parameters = ot.LogNormalMuSigmaOverMu(0.63, 5.24, -0.5)
    myStudy.add("lnmsm_parameters", lnmsm_parameters)
    # WeibullMinMuSigma parameter save
    wms_parameters = ot.WeibullMinMuSigma(1.3, 1.23, -0.5)
    myStudy.add("wms_parameters", wms_parameters)

    # MemoizeFunction
    f = ot.SymbolicFunction(["x1", "x2"], ["x1*x2"])
    memoize = ot.MemoizeFunction(f)
    memoize([5, 6])
    myStudy.add("memoize", memoize)

    # print ('Study = ' , myStudy)
    myStudy.save()

    # Create a new Study Object
    myStudy = ot.Study()
    myStudy.setStorageManager(ot.XMLStorageManager(fileName))

    myStudy.load()
    # print 'loaded Study = ' , myStudy

    # MemoizeFunction
    memoize = ot.MemoizeFunction()
    myStudy.fillObject("memoize", memoize)
    print("memoize = ", repr(memoize))
    memoize([5, 6])
    print("memoize.getCacheHits()=", memoize.getCacheHits())

    # Create a Point from the one stored in the Study
    point = ot.Point()
    myStudy.fillObject("point", point)

    print("point = ", repr(point))

    # Create a Sample from the one stored in the Study
    sample = ot.Sample()
    myStudy.fillObject("mySample", sample)

    print("sample = ", repr(sample))

    # Create a Matrix from the one stored in the Study
    matrix = ot.Matrix()
    myStudy.fillObject("m", matrix)

    print("matrix = ", repr(matrix))

    # Create a Simulation::Result from the one stored in the Study
    simulationResult = ot.ProbabilitySimulationResult()
    myStudy.fillObject("simulationResult", simulationResult)

    print("simulation result = ", simulationResult)

    for cName in cNameList:
        otClass = getattr(ot, cName)
        instance = otClass()
        saved = repr(instance)
        myStudy.fillObject(cName, instance)
        print("--", cName, instance)
        loaded = repr(instance)
        if saved != loaded:
            print("saved=", saved)
            print("loaded=", loaded)

    # Create a Beta distribution from the one stored in the Study
    beta = ot.Beta()
    myStudy.fillObject("beta", beta)

    print("beta = ", beta)

    randomGeneratorState = ot.RandomGeneratorState()

    myStudy.fillObject("randomGeneratorState", randomGeneratorState)

    print("randomGeneratorState = ", randomGeneratorState)

    # Create an analytical Function from the one stored in the
    # Study
    analytical = ot.Function()
    myStudy.fillObject("analytical", analytical)

    print("analytical = ", analytical)
    print("analytical.outputDescription=", analytical.getOutputDescription())

    # Create a GeneralLinearModelResult from the one stored in the Study
    generalizedLinearModelResult = ot.GeneralLinearModelResult()
    myStudy.fillObject("generalizedLinearModelResult", generalizedLinearModelResult)

    print("generalizedLinearModelResult = ", generalizedLinearModelResult)

    # KDTree
    kDTree = ot.KDTree()
    myStudy.fillObject("kDTree", kDTree)

    # ArcsineMuSigma parameter loading
    ams_parameters = ot.ArcsineMuSigma()
    myStudy.fillObject("ams_parameters", ams_parameters)
    # BetaMuSigma parameter loading
    bms_parameters = ot.BetaMuSigma()
    myStudy.fillObject("bms_parameters", bms_parameters)
    # GammaMuSigma parameter loading
    gmms_parameters = ot.GammaMuSigma()
    myStudy.fillObject("gmms_parameters", gmms_parameters)
    # GumbelMuSigma parameter loading
    gms_parameters = ot.GumbelMuSigma()
    myStudy.fillObject("gms_parameters", gms_parameters)
    # LogNormalMuSigma parameter loading
    lnms_parameters = ot.LogNormalMuSigma()
    myStudy.fillObject("lnms_parameters", lnms_parameters)
    # LogNormalMuSigmaOverMu parameter loading
    lnmsm_parameters = ot.LogNormalMuSigmaOverMu()
    myStudy.fillObject("lnmsm_parameters", lnmsm_parameters)
    # WeibullMinMuSigma parameter loading
    wms_parameters = ot.WeibullMinMuSigma()
    myStudy.fillObject("wms_parameters", wms_parameters)

    # test nan/inf
    myStudy = ot.Study(fileName)
    point = ot.Point([float(x) for x in ["nan", "inf", "-inf"]])
    myStudy.add("point", point)
    myStudy.save()
    myStudy2 = ot.Study(fileName)
    myStudy2.load()
    point2 = ot.Point()
    myStudy2.fillObject("point", point2)
    for j in range(len(point2)):
        print(
            "j=",
            j,
            "isnormal=",
            m.isfinite(point2[j]),
            "isnan=",
            m.isnan(point2[j]),
            "isinf=",
            m.isinf(point2[j]),
        )

    # test ComposedDistribution catalog alias
    study = ot.Study()
    study.setStorageManager(ot.XMLStorageManager(fileName))
    distribution = ot.JointDistribution([ot.Uniform()] * 3)
    study.add("distribution", distribution)
    study.save()
    # rewrite xml with the old name
    with open(fileName) as f:
        lines = f.readlines()
    with open(fileName, "w") as f:
        for line in lines:
            f.write(
                line.replace(
                    r'class="JointDistribution"', r'class="ComposedDistribution"'
                )
            )
    study2 = ot.Study(fileName)
    study2.load()
    distribution2 = ot.Distribution()
    study2.fillObject("distribution", distribution2)
    assert distribution2.getDimension() == 3

    # cleanup
    os.remove(fileName)
except Exception:
    import os
    import traceback

    traceback.print_exc()
    os._exit(1)



#! /usr/bin/env python

import openturns as ot
import openturns.testing as ott


mesh = ot.RegularGrid(0.0, 1.0, 4)
values = ot.Sample([[0.5], [1.5], [1.0], [-0.5]])
field = ot.Field(mesh, values)
evaluation = ot.P1LagrangeEvaluation(field)
x = [2.3]
y = evaluation(x)
print(y)
ott.assert_almost_equal(y, [0.55])


# Learning sample on meshD
mesher = ot.IntervalMesher([7, 7])
lowerbound = [-1.0, -1.0]
upperBound = [1, 1]
interval = ot.Interval(lowerbound, upperBound)
meshD = mesher.build(interval)
sample = ot.ProcessSample(meshD, 10, 1)
field = ot.Field(meshD, 1)
for k in range(sample.getSize()):
    field.setValues(ot.Normal().getSample(64))
    sample[k] = field
lagrange = ot.P1LagrangeEvaluation(sample)
# New mesh
mesh = ot.Mesh(
    ot.MonteCarloExperiment(
        ot.JointDistribution([ot.Uniform(-1, 1)] * 2), 200
    ).generate()
)

point = mesh.getVertices()[0]
y = lagrange(point)
print(y)
index = lagrange.getEnclosingSimplexAlgorithm().query(point)
print(index)
assert index == 12, "wrong index"



#!/usr/bin/env python

"""
Calibration without observed inputs
===================================
"""
# %%
# The goal of this example is to present the calibration of a parametric model which
# does not have observed inputs.
# We are going to see how to use the :class:`~openturns.Sample` class
# and create an empty sample.
# Indeed, this is mandatory in order to fit within the calibration
# framework that is used in the library.
# In this example there are, however, several outputs to be calibrated.

import openturns as ot
import openturns.viewer as otv

# %%
# The vector of parameters is
# :math:`\boldsymbol{\theta} = (a, b, c)^T \in \mathbb{R}^3`.
# This model is linear in :math:`\boldsymbol{\theta}` and identifiable.
# It is derived from the equation:
#
# .. math::
#     y(x) = a + b x + c x^2
#
# at :math:`x=-1.0, -0.6, -0.2, 0.2, 0.6, 1.0`.
# In the model, however, the abscissas are fixed and constant.
# Therefore, the parametric model has 3 parameters, no input and 6 outputs
# :math:`y_1, ..., y_6`.
# This produces a set of 5 observations for each output, leading to a total
# of 5 (observations per output) * 6 (outputs) = 30 observations.

# %%
g = ot.SymbolicFunction(
    ["a", "b", "c"],
    [
        "a +  -1.0  * b +  1.0  * c",
        "a +  -0.6  * b +  0.36  * c",
        "a +  -0.2  * b +  0.04  * c",
        "a +  0.2  * b +  0.04  * c",
        "a +  0.6  * b +  0.36  * c",
        "a +  1.0  * b +  1.0  * c",
    ],
)
outputDimension = g.getOutputDimension()
print(outputDimension)

# %%
# We set the true value of the parameters.

# %%
trueParameter = ot.Point([12.0, 7.0, -8.0])
print(trueParameter)

# %%
# In order to generate the observed outputs, we create a distribution
# in dimension 3, with Dirac (i.e. constant) marginals.

# %%
inputRandomVector = ot.JointDistribution([ot.Dirac(theta) for theta in trueParameter])

# %%
# The candidate value is chosen to be different from the true parameter value.

# %%
candidate = ot.Point([8.0, 9.0, -6.0])
calibratedIndices = [0, 1, 2]
model = ot.ParametricFunction(g, calibratedIndices, candidate)

# %%
# We consider a multivariate Gaussian noise with zero mean, standard deviation
# equal to 0.05 and independent copula.
# The independent copula implies that the errors of the 6 outputs are independent.

# %%
outputObservationNoiseSigma = 1.0
meanNoise = ot.Point(outputDimension)
covarianceNoise = [outputObservationNoiseSigma] * outputDimension
R = ot.IdentityMatrix(outputDimension)
observationOutputNoise = ot.Normal(meanNoise, covarianceNoise, R)
print(observationOutputNoise)

# %%
# Finally, we generate the outputs by evaluating the exact outputs of the
# function and adding the noise.
# We use a sample with size equal to 5.

# %%
size = 5
# Generate exact outputs
inputSample = inputRandomVector.getSample(size)
outputStress = g(inputSample)
# Add noise
sampleNoise = observationOutputNoise.getSample(size)
outputObservations = outputStress + sampleNoise

# %%
# Now is the important part of this script: there are no input observations.
# This is why we create a sample of dimension equal to 0.

# %%
inputObservations = ot.Sample(0, 0)

# %%
# We are now ready to perform the calibration.

# %%
algo = ot.LinearLeastSquaresCalibration(
    model, inputObservations, outputObservations, candidate, "SVD"
)
algo.run()
calibrationResult = algo.getResult()
parameterMAP = calibrationResult.getParameterMAP()
print(parameterMAP)


# %%
# We observe that the estimated parameter is relatively close to
# the true parameter value.

# %%
print(parameterMAP - trueParameter)


# %%
# Graphical validation
# --------------------
#
# We now validate the calculation by drawing the exact function and compare
# it to the function with estimated parameters.

# %%
fullModel = ot.SymbolicFunction(["a", "b", "c", "x"], ["a + b * x + c * x^2"])
parameterIndices = [0, 1, 2]
trueFunction = ot.ParametricFunction(fullModel, parameterIndices, trueParameter)
print(trueFunction)
beforeCalibrationFunction = ot.ParametricFunction(
    fullModel, parameterIndices, candidate
)
print(beforeCalibrationFunction)
calibratedFunction = ot.ParametricFunction(fullModel, parameterIndices, parameterMAP)
print(calibratedFunction)

# %%
# In order to validate the calibration, we compare the parametric function
# with true parameters at given abscissas with the parametric function
# with calibrated parameters.

# %%
abscissas = [-1.0, -0.6, -0.2, 0.2, 0.6, 1.0]
xmin = min(abscissas)
xmax = max(abscissas)

npoints = 50
palette = ot.Drawable.BuildDefaultPalette(4)
graph = ot.Graph("Calibration without observations", "x", "y")
graph.setLegendPosition("lower right")
curve = trueFunction.draw(xmin, xmax, npoints).getDrawable(0)
curve.setLineStyle("dotted")
curve.setLegend("True model")
curve.setColor(palette[0])
graph.add(curve)
# Before calibration
curve = beforeCalibrationFunction.draw(xmin, xmax, npoints).getDrawable(0)
curve.setLegend("Model before calibration")
curve.setColor(palette[1])
curve.setLineStyle(ot.ResourceMap.GetAsString("CalibrationResult-PriorLineStyle"))
curve
graph.add(curve)
# After calibration
curve = calibratedFunction.draw(xmin, xmax, npoints).getDrawable(0)
curve.setLegend("Model after calibration")
curve.setColor(palette[2])
curve.setLineStyle(ot.ResourceMap.GetAsString("CalibrationResult-PosteriorLineStyle"))
graph.add(curve)
# Observations
for i in range(outputDimension):
    cloud = ot.Cloud(ot.Sample([[abscissas[i]]] * size), outputObservations[:, i])
    cloud.setColor(palette[3])
    cloud.setPointStyle(
        ot.ResourceMap.GetAsString("CalibrationResult-ObservationPointStyle")
    )
    if i == 0:
        cloud.setLegend("Observations")
    graph.add(cloud)
view = otv.View(graph)


# %%
# We notice that the calibration produces a good fit to the data.
# Still, we notice a small discrepancy between the true mode and the model
# after calibration, but this discrepancy is very small.
# Since the model is linear with respect to the parameters :math:`a`, :math:`b`, :math:`c`,
# the :class:`~openturns.LinearLeastSquares` class performs well.
# If the model were non linear w.r.t. :math:`a`, :math:`b`, :math:`c`, then the linear least
# squares method would not work that well and the parameters would be estimated
# with less accuracy.

otv.View.ShowAll()



"""
Create a customized distribution or copula
==========================================
"""

# %%
# In this example, we show how to create our own distribution or copula.
#
# We use a :class:`~openturns.PythonDistribution` that overloads all the methods of the
# :class:`~openturns.Distribution` class.
# It is not necessary to override all the methods: only the *computeCDF()* and *getRange()* methods
# must be overridden.
# All the methods which are not overridden  are inherited from
# :class:`~openturns.Distribution`: they are numerically derived from the *computeCDF()*
# method, using some keys of the :class:`~openturns.ResourceMap` class.
#
# To create a copula, the method *isCopula()* has to be overridden and return *True*.
#
# Then an instance of the new class can be passed on into a :class:`~openturns.Distribution` class.

# %%
import openturns as ot
import openturns.testing as ott
import openturns.viewer as otv
import math as m
import warnings

warnings.filterwarnings("ignore")

# %%
# First inherit :class:`~openturns.PythonDistribution`:


class UniformNdPy(ot.PythonDistribution):
    def __init__(self, a=[0.0], b=[1.0]):
        super().__init__(len(a))
        if len(a) != len(b):
            raise ValueError("Invalid bounds")
        for i in range(len(a)):
            if a[i] > b[i]:
                raise ValueError("Invalid bounds")
        self.a = a
        self.b = b
        self.factor = 1.0
        for i in range(len(a)):
            self.factor *= b[i] - a[i]

    def getRange(self):
        return ot.Interval(self.a, self.b, [True] * len(self.a), [True] * len(self.a))

    def getRealization(self):
        X = []
        for i in range(len(self.a)):
            X.append(
                self.a[i] + (self.b[i] - self.a[i]) * ot.RandomGenerator.Generate()
            )
        return X

    def getSample(self, size):
        X = []
        for i in range(size):
            X.append(self.getRealization())
        return X

    def computeCDF(self, X):
        prod = 1.0
        for i in range(len(self.a)):
            if X[i] < self.a[i]:
                return 0.0
            prod *= min(self.b[i], X[i]) - self.a[i]
        return prod / self.factor

    def computePDF(self, X):
        for i in range(len(self.a)):
            if X[i] < self.a[i]:
                return 0.0
            if X[i] > self.b[i]:
                return 0.0
        return 1.0 / self.factor

    def getMean(self):
        mu = []
        for i in range(len(self.a)):
            mu.append(0.5 * (self.a[i] + self.b[i]))
        return mu

    def getStandardDeviation(self):
        stdev = []
        for i in range(len(self.a)):
            stdev.append((self.b[i] - self.a[i]) / m.sqrt(12.0))
        return stdev

    def getSkewness(self):
        return [0.0] * len(self.a)

    def getKurtosis(self):
        return [1.8] * len(self.a)

    def getMoment(self, n):
        return [-0.1 * n] * len(self.a)

    def computeCharacteristicFunction(self, x):
        if len(self.a) > 1:
            raise ValueError("dim>1")
        ax = self.a[0] * x
        bx = self.b[0] * x
        return (m.sin(bx) - m.sin(ax) + 1j * (m.cos(ax) - m.cos(bx))) / (bx - ax)

    def isElliptical(self):
        return (len(self.a) == 1) and (self.a[0] == -self.b[0])

    def isCopula(self):
        for i in range(len(self.a)):
            if self.a[i] != 0.0:
                return False
            if self.b[i] != 1.0:
                return False
        return True

    def getMarginal(self, indices):
        subA = []
        subB = []
        for i in indices:
            subA.append(self.a[i])
            subB.append(self.b[i])
        py_dist = UniformNdPy(subA, subB)
        return ot.Distribution(py_dist)

    def computeQuantile(self, prob, tail=False):
        p = 1.0 - prob if tail else prob
        quantile = list(self.a)
        for i in range(len(self.a)):
            quantile[i] += p * (self.b[i] - self.a[i])
        return quantile


# %%
# Now instantiate the distribution:
distribution = ot.Distribution(UniformNdPy([5, 6], [7, 9]))

# %%
# And plot the CDF:
graph = distribution.drawCDF()
view = otv.View(graph)

# %%
# We can easily generate sample:
distribution.getSample(5)

# %%
# or compute the mean:
distribution.getMean()

# %%
# Also we can compute the probability contained in an interval:
distribution.computeProbability(ot.Interval([5.5, 6], [8.5, 9]))

# %%
# We can validate the distribution with :class:`openturns.testing.DistributionValidation`
# It automatically checks the consistency of most services and allows one to check for errors.
checker = ott.DistributionValidation(distribution)
checker.run()

# %%
otv.View.ShowAll()



"""
Compare unconditional and conditional histograms
================================================
"""

# %%
# In this example, we compare unconditional and conditional histograms for a
# simulation.
# We consider the :ref:`flooding model<use-case-flood-model>`.
# Let :math:`g` be a function which takes four inputs :math:`Q`, :math:`K_s`,
# :math:`Z_v` and :math:`Z_m` and returns one output :math:`S`.
#
# We first consider the (unconditional) distribution of the input :math:`Q`.
#
# Let :math:`t` be a given threshold on the output :math:`S`: we consider
# the event :math:`S > t`.
# Then we consider the conditional distribution of the input :math:`Q` given
# that :math:`S > t` that is to say :math:`Q|S > t`.
#
# If these two distributions are significantly different, we conclude that
# the input :math:`Q` has an impact on the event :math:`S > t`.
#
# In order to approximate the distribution of the output :math:`S`,
# we perform a Monte-Carlo simulation with size 500.
# The threshold :math:`t` is chosen as the 90% quantile of the empirical
# distribution of :math:`S`.
# In this example, the distribution is approximated by its empirical histogram
# (but this could be done with another distribution approximation as well,
# such as kernel smoothing for example).

# %%
import numpy as np
from openturns.usecases import flood_model
import openturns as ot
import openturns.viewer as otv


# %%
# We use the `FloodModel` data class that contains all the case parameters.
fm = flood_model.FloodModel()


# %%
# Create an input sample from the joint `distribution` defined in the data class.
# We build an output sample by taking the image by the `model`.

# %%
size = 500
inputSample = fm.distribution.getSample(size)
inputSample[:5]

# %%
outputSample = fm.model(inputSample)
outputSample[:5]

# %%
# Merge the input and output samples into a single sample.

# %%
sample = ot.Sample(inputSample)
sample.stack(outputSample)
sample[0:5]

# %%
# Extract the first column of `inputSample` into the sample of the flowrates
# :math:`Q`.

# %%
sampleQ = inputSample[:, 0]

# %%
# The next cell defines a function that computes the conditional sample of
# a component given that the a marginal (defined by its index `criteriaComponent`)
# exceeds a given threshold, defined by its quantile level.


def computeConditionnedSample(
    sample, alpha=0.9, criteriaComponent=None, selectedComponent=0
):
    """
    Return values from the selectedComponent-th component of the sample.
    Selects the values according to the alpha-level quantile of
    the criteriaComponent-th component of the sample.
    """
    dim = sample.getDimension()
    if criteriaComponent is None:
        criteriaComponent = dim - 1
    sortedSample = sample.sortAccordingToAComponent(criteriaComponent)
    quantiles = sortedSample.computeQuantilePerComponent(alpha)
    quantileValue = quantiles[criteriaComponent]
    sortedSampleCriteria = sortedSample[:, criteriaComponent]
    indices = np.where(np.array(sortedSampleCriteria.asPoint()) > quantileValue)[0]
    conditionnedSortedSample = sortedSample[int(indices[0]) :, selectedComponent]
    return conditionnedSortedSample


# %%
# Create an histogram for the unconditional flowrates.

# %%
numberOfBins = 10
histogram = ot.HistogramFactory().buildAsHistogram(sampleQ, numberOfBins)

# %%
# Extract the sub-sample of the input flowrates `Q` which leads to large values of the output `S`.

# %%
# Search the index of the marginal `S` in the columns of the sample.
criteriaComponent = list(sample.getDescription()).index("S")
criteriaComponent

# %%
alpha = 0.9
selectedComponent = 0
conditionnedSampleQ = computeConditionnedSample(
    sample, alpha, criteriaComponent, selectedComponent
)

# %%
# We could as well use:
#
# .. code-block::
#
#     # conditionnedHistogram = ot.HistogramFactory().buildAsHistogram(conditionnedSampleQ)
#
# but this creates an histogram with new classes, corresponding
# to `conditionnedSampleQ`.
# We want to use exactly the same classes as the full sample,
# so that the two histograms match.

# %%
first = histogram.getFirst()
width = histogram.getWidth()
conditionnedHistogram = ot.HistogramFactory().buildAsHistogram(
    conditionnedSampleQ, first, width
)

# %%
# Then creates a graphics with the unconditional and the conditional histograms.

# %%
graph = histogram.drawPDF()
graph.setLegends(["Q"])
#
graphConditionnalQ = conditionnedHistogram.drawPDF()
graphConditionnalQ.setLegends([r"$Q | S > S_{%s}$" % (alpha)])
graph.add(graphConditionnalQ)
view = otv.View(graph)


# %%
# Show all the graphs.
view.ShowAll()

# %%
# We see that the two histograms are very different.
# The high values of the input :math:`Q` seem to often lead to a high value of the output :math:`S`.
#
# We could explore this situation further by comparing the unconditional
# distribution of :math:`Q` (which is known in this case) with the conditonal
# distribution of :math:`Q | S > t`, estimated by kernel smoothing.
# This would have the advantage of accuracy, since the kernel smoothing is a
# more accurate approximation of a distribution than the histogram.



import openturns as ot
from matplotlib import pyplot as plt
import openturns.viewer as otv


f = ot.SymbolicFunction(["x"], ["sin(x)"])
a = -2.5
b = 4.5
algo = ot.GaussLegendre([20])
value, nodes = algo.integrateWithNodes(f, ot.Interval(a, b))

g = f.draw(a, b, 512)
lower = ot.Cloud(nodes, ot.Sample(nodes.getSize(), 1))
lower.setColor("magenta")
lower.setPointStyle("circle")
g.add(lower)
g.setTitle(r"GaussLegendre example: $\int_{-5/2}^{9/2}\sin(t)\,dt=$" + str(value[0]))

fig = plt.figure(figsize=(8, 4))
axis = fig.add_subplot(111)
axis.set_xlim(auto=True)
otv.View(g, figure=fig, axes=[axis], add_legend=False)



import openturns as ot
import openturns.viewer as otv
from openturns.usecases import ishigami_function

im = ishigami_function.IshigamiModel()
sampleSize = 500
inputTrain = im.distribution.getSample(sampleSize)
outputTrain = im.model(inputTrain)
multivariateBasis = ot.OrthogonalProductPolynomialFactory([im.X1, im.X2, im.X3])
selectionAlgorithm = ot.PenalizedLeastSquaresAlgorithmFactory()
projectionStrategy = ot.LeastSquaresStrategy(
    inputTrain, outputTrain, selectionAlgorithm
)
totalDegree = 8
enumerateFunction = multivariateBasis.getEnumerateFunction()
basisSize = enumerateFunction.getStrataCumulatedCardinal(totalDegree)
adaptiveStrategy = ot.FixedStrategy(multivariateBasis, basisSize)
chaosalgo = ot.FunctionalChaosAlgorithm(
    inputTrain, outputTrain, im.distribution, adaptiveStrategy, projectionStrategy
)
chaosalgo.run()
result = chaosalgo.getResult()
validation = ot.FunctionalChaosValidation(result)
graph = validation.drawValidation()
otv.View(graph)



"""
Composite random vector
=======================
"""

# %%
# In this example we are going to create a random variable :math:`\vect{Y}` which realizations are the images
# of the realizations of another random vector :math:`\vect{X}` by a function.
#
# .. math::
#     \vect{Y} = f(\vect{X})
#

# %%
import openturns as ot


# %%
# Create a random vector based on a distribution
dist2d = ot.Normal(2)
X = ot.RandomVector(dist2d)

# %%
# Create the function
f = ot.SymbolicFunction(["x1", "x2"], ["x1 + x2", "x1*x2"])
f

# %%
# Create the composite random vector
Y = ot.CompositeRandomVector(f, X)

# %%
Y.getSample(5)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
"""
@author: Antoine Dumas (Phiméca)

Calcul de la variance des estimateurs de Sobol d'openTURNS
"""


import openturns as ot
import numpy as np


def computeSobolVariance(U, psi, size):
    """
    Compute the variance of the estimators

    Parameters
    ----------
    U : sample
        The sample of yA, yB, yE or combination of them, defined according the
        sobol estimators
    psi : Function
        The function that computes the sobol estimates.
    size : int
        The size of the sample.
    """
    mean_psi = psi.gradient(U.computeMean()) * ot.Point(
        1, 1
    )  # to transform into a Point
    variance = mean_psi.dot(U.computeCovariance() * mean_psi) / size
    return variance


class SaltelliSensitivityAlgorithm(ot.SaltelliSensitivityAlgorithm):
    def __init__(self, inputDesign, outputDesign, N):
        super(SaltelliSensitivityAlgorithm, self).__init__(inputDesign, outputDesign, N)
        self.inputDesign = inputDesign
        self.input_dim = inputDesign.getDimension()
        self.output_dim = outputDesign.getDimension()
        self.size = N
        # centrage de l'échantillon de sortie
        self.outputDesign = outputDesign  # - outputDesign.computeMean()[0]

    def computeVariance(self):

        x = ot.Description.BuildDefault(self.output_dim, "X")
        y = ot.Description.BuildDefault(self.output_dim, "Y")
        # in order X0, Y0, X1, Y1, ...
        xy = list(x)
        for i, yy in enumerate(y):
            xy.insert(2 * i + 1, yy)
        # psi  = (x1 + x2 + ...) / (y1 + y2 + ...).
        symbolic_num = ""
        symbolic_denom = ""
        for i in range(self.output_dim):
            symbolic_num += x[i]
            symbolic_denom += y[i]
            if i < self.output_dim - 1:
                symbolic_num += "+"
                symbolic_denom += "+"
        psi_fo = ot.SymbolicFunction(
            xy, ["(" + symbolic_num + ")/(" + symbolic_denom + ")"]
        )
        psi_to = ot.SymbolicFunction(
            xy, ["1 - " + "(" + symbolic_num + ")/(" + symbolic_denom + ")"]
        )

        varianceFO = ot.Point(self.input_dim)
        varianceTO = ot.Point(self.input_dim)
        for p in range(self.input_dim):
            U_fo = ot.Sample(self.size, 0)
            U_to = ot.Sample(self.size, 0)
            for q in range(self.output_dim):

                yA = ot.Sample(self.outputDesign[:, q], 0, self.size)
                yB = ot.Sample(self.outputDesign[:, q], self.size, 2 * self.size)
                yAc = yA - yA.computeMean()[0]
                yBc = yB - yB.computeMean()[0]
                yE = ot.Sample(
                    self.outputDesign[:, q], (2 + p) * self.size, (3 + p) * self.size
                )
                yEc = yE - yE.computeMean()[0]

                # first order
                U_fo.stack(np.array(yBc) * np.array(yEc))
                U_fo.stack(np.array(yAc) ** 2)  # centré dans tous les cas ici

                # total order
                U_to.stack(np.array(yAc) * np.array(yEc))
                U_to.stack(np.array(yAc) ** 2)  # centré dans tous les cas ici

            varianceFO[p] = computeSobolVariance(U_fo, psi_fo, self.size)
            varianceTO[p] = computeSobolVariance(U_to, psi_to, self.size)

        return varianceFO, varianceTO

    def getFirstOrderAsymptoticDistribution(self):
        indicesFO = self.getAggregatedFirstOrderIndices()
        varianceFO, varianceTO = self.computeVariance()
        foDist = ot.DistributionCollection(self.input_dim)
        for p in range(self.input_dim):
            foDist[p] = ot.Normal(indicesFO[p], np.sqrt(varianceFO[p]))
        return foDist

    def getTotalOrderAsymptoticDistribution(self):
        indicesTO = self.getAggregatedTotalOrderIndices()
        varianceFO, varianceTO = self.computeVariance()
        toDist = ot.DistributionCollection(self.input_dim)
        for p in range(self.input_dim):
            toDist[p] = ot.Normal(indicesTO[p], np.sqrt(varianceTO[p]))
        return toDist

    def getAggregatedFirstOrderIndices(self):

        sumVariance = 0
        VarianceI = ot.Point(self.input_dim)
        for q in range(self.output_dim):
            yA = ot.Sample(self.outputDesign[:, q], 0, self.size)
            yAc = yA - yA.computeMean()[0]
            yB = ot.Sample(self.outputDesign[:, q], self.size, 2 * self.size)
            yBc = yB - yB.computeMean()[0]
            sumVariance += yA.computeVariance()[0]

            # FOindices = ot.Point(self.input_dim)
            for p in range(self.input_dim):
                yE = ot.Sample(
                    self.outputDesign[:, q], (2 + p) * self.size, (3 + p) * self.size
                )
                yEc = yE - yE.computeMean()[0]

                xc = np.array(yBc) * np.array(yEc)
                VarianceI[p] += np.mean(xc) - np.mean(yAc) * np.mean(yBc)

        FOindices = ot.Point(VarianceI / sumVariance)
        return FOindices


class JansenSensitivityAlgorithm(ot.JansenSensitivityAlgorithm):
    def __init__(self, inputDesign, outputDesign, N):
        super(JansenSensitivityAlgorithm, self).__init__(inputDesign, outputDesign, N)
        self.inputDesign = inputDesign
        self.input_dim = inputDesign.getDimension()
        self.output_dim = outputDesign.getDimension()
        self.size = N
        # centrage de l'échantillon de sortie
        self.outputDesign = outputDesign  # - outputDesign.computeMean()[0]

    def computeVariance(self):

        x = ot.Description.BuildDefault(self.output_dim, "X")
        y = ot.Description.BuildDefault(self.output_dim, "Y")
        # in order X0, Y0, X1, Y1, ...
        xy = list(x)
        for i, yy in enumerate(y):
            xy.insert(2 * i + 1, yy)
        # psi  = 0.5*(x1 + x2 + ...) / (y1 + y2 + ...).
        symbolic_num = ""
        symbolic_denom = ""
        for i in range(self.output_dim):
            symbolic_num += x[i]
            symbolic_denom += y[i]
            if i < self.output_dim - 1:
                symbolic_num += "+"
                symbolic_denom += "+"
        psi_fo = ot.SymbolicFunction(
            xy, ["1 - " + "0.5*(" + symbolic_num + ")/(" + symbolic_denom + ")"]
        )
        psi_to = ot.SymbolicFunction(
            xy, ["0.5*(" + symbolic_num + ")/(" + symbolic_denom + ")"]
        )

        varianceFO = ot.Point(self.input_dim)
        varianceTO = ot.Point(self.input_dim)
        for p in range(self.input_dim):
            U_fo = ot.Sample(self.size, 0)
            U_to = ot.Sample(self.size, 0)

            for q in range(self.output_dim):
                yA = ot.Sample(self.outputDesign[:, q], 0, self.size)
                yB = ot.Sample(self.outputDesign[:, q], self.size, 2 * self.size)
                yAc = yA - yA.computeMean()[0]
                yBc = yB - yB.computeMean()[0]
                yE = ot.Sample(
                    self.outputDesign[:, q], (2 + p) * self.size, (3 + p) * self.size
                )
                yEc = yE - yE.computeMean()[0]

                # First order
                U_fo.stack(np.array(yEc - yBc) ** 2)
                U_fo.stack(np.array(yAc) ** 2)

                # Total order
                U_to.stack(np.array(yAc - yEc) ** 2)
                U_to.stack(np.array(yAc) ** 2)

            varianceFO[p] = computeSobolVariance(U_fo, psi_fo, self.size)
            varianceTO[p] = computeSobolVariance(U_to, psi_to, self.size)

        return varianceFO, varianceTO

    def getFirstOrderAsymptoticDistribution(self):
        indicesFO = self.getAggregatedFirstOrderIndices()
        varianceFO, varianceTO = self.computeVariance()
        foDist = ot.DistributionCollection(self.input_dim)
        for p in range(self.input_dim):
            foDist[p] = ot.Normal(indicesFO[p], np.sqrt(varianceFO[p]))
        return foDist

    def getTotalOrderAsymptoticDistribution(self):
        indicesTO = self.getAggregatedTotalOrderIndices()
        varianceFO, varianceTO = self.computeVariance()
        toDist = ot.DistributionCollection(self.input_dim)
        for p in range(self.input_dim):
            toDist[p] = ot.Normal(indicesTO[p], np.sqrt(varianceTO[p]))
        return toDist


class MauntzKucherenkoSensitivityAlgorithm(ot.MauntzKucherenkoSensitivityAlgorithm):
    def __init__(self, inputDesign, outputDesign, N):
        super(MauntzKucherenkoSensitivityAlgorithm, self).__init__(
            inputDesign, outputDesign, N
        )
        self.inputDesign = inputDesign
        self.input_dim = inputDesign.getDimension()
        self.output_dim = outputDesign.getDimension()
        self.size = N
        # centrage de l'échantillon de sortie
        self.outputDesign = outputDesign  # - outputDesign.computeMean()[0]

    def computeVariance(self):

        x = ot.Description.BuildDefault(self.output_dim, "X")
        y = ot.Description.BuildDefault(self.output_dim, "Y")
        # in order X0, Y0, X1, Y1, ...
        xy = list(x)
        for i, yy in enumerate(y):
            xy.insert(2 * i + 1, yy)
        # psi  = (x1 + x2 + ...) / (y1 + y2 + ...).
        symbolic_num = ""
        symbolic_denom = ""
        for i in range(self.output_dim):
            symbolic_num += x[i]
            symbolic_denom += y[i]
            if i < self.output_dim - 1:
                symbolic_num += "+"
                symbolic_denom += "+"
        psi_fo = ot.SymbolicFunction(
            xy, ["(" + symbolic_num + ")/(" + symbolic_denom + ")"]
        )
        psi_to = ot.SymbolicFunction(
            xy, ["1 - " + "(" + symbolic_num + ")/(" + symbolic_denom + ")"]
        )

        varianceFO = ot.Point(self.input_dim)
        varianceTO = ot.Point(self.input_dim)
        for p in range(self.input_dim):
            U_fo = ot.Sample(self.size, 0)
            U_to = ot.Sample(self.size, 0)

            for q in range(self.output_dim):
                yA = ot.Sample(self.outputDesign[:, q], 0, self.size)
                yB = ot.Sample(self.outputDesign[:, q], self.size, 2 * self.size)
                yAc = yA - yA.computeMean()[0]
                yBc = yB - yB.computeMean()[0]
                yE = ot.Sample(
                    self.outputDesign[:, q], (2 + p) * self.size, (3 + p) * self.size
                )
                yEc = yE - yE.computeMean()[0]

                # First order
                U_fo.stack(np.array(yBc) * np.array(yEc - yAc))
                U_fo.stack(np.array(yAc) ** 2)
                # psi_fo = ot.SymbolicFunction(['x', 'y'], ['x / y'])

                # Total order
                U_to.stack(np.array(yAc) * np.array(yEc - yAc))
                U_to.stack(np.array(yAc) ** 2)
                # psi_to = ot.SymbolicFunction(['x', 'y'], ['1 - x / y'])

            varianceFO[p] += computeSobolVariance(U_fo, psi_fo, self.size)
            varianceTO[p] += computeSobolVariance(U_to, psi_to, self.size)

        return varianceFO, varianceTO

    def getFirstOrderAsymptoticDistribution(self):
        indicesFO = self.getAggregatedFirstOrderIndices()
        varianceFO, varianceTO = self.computeVariance()
        foDist = ot.DistributionCollection(self.input_dim)
        for p in range(self.input_dim):
            foDist[p] = ot.Normal(indicesFO[p], np.sqrt(varianceFO[p]))
        return foDist

    def getTotalOrderAsymptoticDistribution(self):
        indicesTO = self.getAggregatedTotalOrderIndices()
        varianceFO, varianceTO = self.computeVariance()
        toDist = ot.DistributionCollection(self.input_dim)
        for p in range(self.input_dim):
            toDist[p] = ot.Normal(indicesTO[p], np.sqrt(varianceTO[p]))
        return toDist


class MartinezSensitivityAlgorithm(ot.MartinezSensitivityAlgorithm):
    def __init__(self, inputDesign, outputDesign, N):
        super(MartinezSensitivityAlgorithm, self).__init__(inputDesign, outputDesign, N)
        self.inputDesign = inputDesign
        self.input_dim = inputDesign.getDimension()
        self.output_dim = outputDesign.getDimension()
        self.size = N
        # centrage de l'échantillon de sortie
        self.outputDesign = outputDesign  # - outputDesign.computeMean()[0]

    def computeVariance(self):

        x = ot.Description.BuildDefault(self.output_dim, "X")
        y = ot.Description.BuildDefault(self.output_dim, "Y")
        z = ot.Description.BuildDefault(self.output_dim, "Z")
        # in order X0, Y0, Z0, X1, Y1, Z1, ...
        xyz = list(x)
        for i in range(self.output_dim):
            xyz.insert(3 * i + 1, y[i])
            xyz.insert(3 * i + 2, z[i])

        # psi_fo  = (x1 + x2 + ...) / (sqrt(y1*z1) + sqrt(y2*z2) + ...).
        symbolic_num = ""
        symbolic_denom = ""
        for i in range(self.output_dim):
            symbolic_num += x[i]
            symbolic_denom += "sqrt(" + y[i] + "*" + z[i] + ")"
            if i < self.output_dim - 1:
                symbolic_num += "+"
                symbolic_denom += "+"
        psi_fo = ot.SymbolicFunction(
            xyz, ["(" + symbolic_num + ")/(" + symbolic_denom + ")"]
        )
        psi_to = ot.SymbolicFunction(
            xyz, ["1 - " + "(" + symbolic_num + ")/(" + symbolic_denom + ")"]
        )

        varianceFO = ot.Point(self.input_dim)
        varianceTO = ot.Point(self.input_dim)
        for p in range(self.input_dim):
            U_fo = ot.Sample(self.size, 0)
            U_to = ot.Sample(self.size, 0)

            for q in range(self.output_dim):
                yA = ot.Sample(self.outputDesign[:, q], 0, self.size)
                yB = ot.Sample(self.outputDesign[:, q], self.size, 2 * self.size)
                yAc = yA - yA.computeMean()[0]
                yBc = yB - yB.computeMean()[0]
                yE = ot.Sample(
                    self.outputDesign[:, q], (2 + p) * self.size, (3 + p) * self.size
                )
                yEc = yE - yE.computeMean()[0]

                # First order
                U_fo.stack(np.array(yBc) * np.array(yEc))
                U_fo.stack(np.array(yBc) ** 2)
                U_fo.stack(np.array(yEc) ** 2)
                # psi_fo = ot.SymbolicFunction(['x', 'y', 'z'], ['x / sqrt(y*z)'])

                # Total order
                U_to.stack(np.array(yAc) * np.array(yEc))
                U_to.stack(np.array(yAc) ** 2)
                U_to.stack(np.array(yEc) ** 2)
                # psi_to = ot.SymbolicFunction(['x', 'y', 'z'], ['1 - x / sqrt(y*z)'])

            varianceFO[p] += computeSobolVariance(U_fo, psi_fo, self.size)
            varianceTO[p] += computeSobolVariance(U_to, psi_to, self.size)

        return varianceFO, varianceTO

    def getFirstOrderAsymptoticDistribution(self):
        indicesFO = self.getAggregatedFirstOrderIndices()
        varianceFO, varianceTO = self.computeVariance()
        foDist = ot.DistributionCollection(self.input_dim)
        for p in range(self.input_dim):
            foDist[p] = ot.Normal(indicesFO[p], np.sqrt(varianceFO[p]))
        return foDist

    def getTotalOrderAsymptoticDistribution(self):
        indicesTO = self.getAggregatedTotalOrderIndices()
        varianceFO, varianceTO = self.computeVariance()
        toDist = ot.DistributionCollection(self.input_dim)
        for p in range(self.input_dim):
            toDist[p] = ot.Normal(indicesTO[p], np.sqrt(varianceTO[p]))
        return toDist



#! /usr/bin/env python

import openturns as ot

levels = [4] * 2
experiment = ot.Box(levels)
print("experiment = ", experiment)
sample = experiment.generate()
print("sample = ", repr(sample))

experiment = ot.Box(levels, ot.Interval([7.0, 8.0], [9.0, 16.0]))
print("experiment = ", experiment)
sample = experiment.generate()
print("sample = ", repr(sample))



import openturns as ot
from matplotlib import pyplot as plt
import openturns.viewer as otv

distribution = ot.NormalGamma()

distribution.setDescription(["$X$", "$Y$"])
pdf_graph = distribution.drawPDF()
pdf_contour = pdf_graph.getDrawable(0).getImplementation()
pdf_contour.setColorBarPosition("")
pdf_contour.setColorMapNorm("rank")
pdf_graph.setDrawable(0, pdf_contour)
pdf_graph.setTitle(str(distribution))
fig = plt.figure(figsize=(5, 5))
pdf_axis = fig.add_subplot(111)
otv.View(pdf_graph, figure=fig, axes=[pdf_axis], add_legend=False)



"""
Time variant system reliability problem
=======================================
"""

# %%

# %%
# The objective is to evaluate the outcrossing rate from a safe to a failure domain in a time variant reliability problem.
#
# We consider the following limit state function, defined as the difference between a degrading resistance :math:`r(t) = R - bt`  and a time-varying load :math:`S(t)`:
#
# .. math::
#   \begin{align*}
#   g(t)= r(t) - S(t) = R - bt - S(t) \quad \forall t \in [0,T]
#   \end{align*}
#
# The failure domaine is defined by:
#
# .. math::
#    g(t) \leq 0
#
#
# which means that the resistance at :math:`t` is less than the stress at :math:`t`.
#
#
# We propose the following probabilistic model:
#
# - :math:`R` is the initial resistance, and :math:`R \sim \mathcal{N}(\mu_R, \sigma_R)`;
# - :math:`b` is the deterioration rate of the resistance; it is deterministic;
# - :math:`S(\omega,t)` is the time-varying stress, which is modeled by a stationary Gaussian process of mean value :math:`\mu_S`,
#   standard deviation :math:`\sigma_S` and a squared exponential covariance model :math:`C(s,t)`.
#
#
# The outcrossing rate from the safe to the failure domain at instant :math:`t` is defined by:
#
# .. math::
#    \nu^+(t) = \lim_{\Delta t \rightarrow 0+} \dfrac{\mathbb{P}\{ g(t) \ge 0 \cap g(t+\Delta t) \leq 0\} }{\Delta t}
#
#
# For each :math:`t`, we note the random variable  :math:`Z_t = g(t)`.
#
# To evaluate :math:`\nu^+(t)`, we need to consider the bivariate random vector :math:`(Z_t, Z_{t+\Delta t})`.
#
# The event :math:`\{ g(t) \geq 0 \cap g(t+\Delta t) <0\}` writes as the intersection of both events :
#
# - :math:`\mathcal{E}_t^1 = \{   Z_t \geq 0\}` and
# - :math:`\mathcal{E}_t^2 = \{   Z_{t+\Delta t} \leq 0\}`.
#
# The objective is to evaluate:
#
# .. math::
#    \mathbb{P}\{\mathcal{E}_t^1 \cap \mathcal{E}_t^2\} \quad \forall t \in [0,T]
#

# %%
# 1. Define some useful functions
# -------------------------------

# %%
# We define the bivariate random vector :math:`Y_t = (bt + S_t, b(t+\Delta t) + S_{t+\Delta t})` where :math:`S_t = S(., t)`.
# Here, :math:`Y_t` is a bivariate Normal random vector:
#
# - with mean :math:`[bt, b(t+\delta t)]` and
# - with covariance matrix :math:`\Sigma` defined by:
#
# .. math::
#   \begin{align*}
#   \Sigma = \left(
#   \begin{array}{cc}
#   C(t, t) & C(t, t+\Delta t) \\
#   C(t, t+\Delta t) & C(t+\Delta t, t+\Delta t)
#   \end{array}
#   \right)
#   \end{align*}
#
# This function buils :math:`Y_t =(Y_t^1, Y_t^2)`.

# %%
from math import sqrt
import openturns.viewer as otv
import openturns as ot


def buildNormal(b, t, mu_S, covariance, delta_t=1e-5):
    sigma = ot.CovarianceMatrix(2)
    sigma[0, 0] = covariance(t, t)[0, 0]
    sigma[0, 1] = covariance(t, t + delta_t)[0, 0]
    sigma[1, 1] = covariance(t + delta_t, t + delta_t)[0, 0]
    return ot.Normal([b * t + mu_S, b * (t + delta_t) + mu_S], sigma)


# %%
# This function creates the trivariate random vector :math:`(R, Y_t^1, Y_t^2)` where :math:`R` is independent from :math:`(Y_t^1, Y_t^2)`.
# We need to create this random vector because both events  :math:`\mathcal{E}_t^1` and :math:`\mathcal{E}_t^2` must be defined from the same random vector!


# %%
def buildCrossing(b, t, mu_S, covariance, R, delta_t=1e-5):
    normal = buildNormal(b, t, mu_S, covariance, delta_t)
    return ot.BlockIndependentDistribution([R, normal])


# %%
# This function evaluates the probability using the Monte Carlo sampling. It defines the intersection event :math:`\mathcal{E}_t^1 \cap \mathcal{E}_t^2`.


# %%
def getXEvent(b, t, mu_S, covariance, R, delta_t):
    full = buildCrossing(b, t, mu_S, covariance, R, delta_t)
    X = ot.RandomVector(full)
    f1 = ot.SymbolicFunction(["R", "X1", "X2"], ["X1 - R"])
    e1 = ot.ThresholdEvent(ot.CompositeRandomVector(f1, X), ot.Less(), 0.0)
    f2 = ot.SymbolicFunction(["R", "X1", "X2"], ["X2 - R"])
    e2 = ot.ThresholdEvent(ot.CompositeRandomVector(f2, X), ot.GreaterOrEqual(), 0.0)
    event = ot.IntersectionEvent([e1, e2])
    return X, event


# %%
def computeCrossingProbability_MonteCarlo(
    b, t, mu_S, covariance, R, delta_t, n_block, n_iter, CoV
):
    X, event = getXEvent(b, t, mu_S, covariance, R, delta_t)
    algo = ot.ProbabilitySimulationAlgorithm(event, ot.MonteCarloExperiment())
    algo.setBlockSize(n_block)
    algo.setMaximumOuterSampling(n_iter)
    algo.setMaximumCoefficientOfVariation(CoV)
    algo.run()
    return algo.getResult().getProbabilityEstimate() / delta_t


# %%
# This function evaluates the probability using the Low Discrepancy sampling.


# %%
def computeCrossingProbability_QMC(
    b, t, mu_S, covariance, R, delta_t, n_block, n_iter, CoV
):
    X, event = getXEvent(b, t, mu_S, covariance, R, delta_t)
    algo = ot.ProbabilitySimulationAlgorithm(
        event,
        ot.LowDiscrepancyExperiment(ot.SobolSequence(X.getDimension()), n_block, False),
    )
    algo.setBlockSize(n_block)
    algo.setMaximumOuterSampling(n_iter)
    algo.setMaximumCoefficientOfVariation(CoV)
    algo.run()
    return algo.getResult().getProbabilityEstimate() / delta_t


# %%
# This function evaluates the probability using the FORM algorithm for event systems..


# %%
def computeCrossingProbability_FORM(b, t, mu_S, covariance, R, delta_t):
    X, event = getXEvent(b, t, mu_S, covariance, R, delta_t)
    solver = ot.SQP()
    solver.setStartingPoint(X.getMean())
    algo = ot.SystemFORM(solver, event)
    algo.run()
    return algo.getResult().getEventProbability() / delta_t


# %%
# 2. Evaluate the probability
# ---------------------------

# %%

# %%
# First, fix some parameters: :math:`(\mu_R, \sigma_R, \mu_S, \sigma_S, \Delta t, T, b)` and the covariance model which is the Squared Exponential model.
# Be careful to the parameter  :math:`\Delta t` which is of great importance: if it is too small, the simulation methods have problems to converge
# because the correlation rate is too high between the instants :math:`t` and :math:`t+\Delta t`.
# We advice to take :math:`\Delta t \simeq 10^{-1}`.
#

# %%
mu_S = 3.0
sigma_S = 0.5
ll = 10

b = 0.01

mu_R = 5.0
sigma_R = 0.3
R = ot.Normal(mu_R, sigma_R)

covariance = ot.SquaredExponential([ll / sqrt(2)], [sigma_S])

t0 = 0.0
t1 = 50.0
N = 26

# Get all the time steps t
times = ot.RegularGrid(t0, (t1 - t0) / (N - 1.0), N).getVertices()

delta_t = 1e-1

# %%
# Use all the methods previously described:
#
# - Monte Carlo: values in values_MC
# - Low discrepancy suites: values in values_QMC
# - FORM: values in values_FORM
#

# %%
values_MC = list()
values_QMC = list()
values_FORM = list()

for tick in times:
    values_MC.append(
        computeCrossingProbability_MonteCarlo(
            b, tick[0], mu_S, covariance, R, delta_t, 2**12, 2**3, 1e-2
        )
    )
    values_QMC.append(
        computeCrossingProbability_QMC(
            b, tick[0], mu_S, covariance, R, delta_t, 2**12, 2**3, 1e-2
        )
    )
    values_FORM.append(
        computeCrossingProbability_FORM(b, tick[0], mu_S, covariance, R, delta_t)
    )

# %%
print("Values MC = ", values_MC)
print("Values QMC = ", values_QMC)
print("Values FORM = ", values_FORM)

# %%
# Draw the graphs!

# %%
g = ot.Graph()
g.setAxes(True)
g.setGrid(True)
c = ot.Curve(times, [[p] for p in values_MC])
g.add(c)
c = ot.Curve(times, [[p] for p in values_QMC])
g.add(c)
c = ot.Curve(times, [[p] for p in values_FORM])
g.add(c)
g.setLegends(["MC", "QMC", "FORM"])
g.setLegendPosition("upper left")
g.setXTitle("t")
g.setYTitle("Outcrossing rate")
view = otv.View(g)

# %%
# Display all the figures
view.ShowAll()



#! /usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()

print("ResourceMap={")
for key in ot.ResourceMap.GetKeys():
    print("  %s => %s," % (key, ot.ResourceMap.Get(key)))
print("}")
print(
    "Extract from ResourceMap: Cache-MaxSize -> ",
    ot.ResourceMap.Get("Cache-MaxSize"),
)

# check string enum api
ot.ResourceMap.AddAsString("bar", "v")
ot.ResourceMap.AddAsString("foo", "a", ["a", "b"])
assert ot.ResourceMap.HasStringEnum("foo")
assert ot.ResourceMap.GetStringEnum("foo") == ("a", "b")
ot.ResourceMap.SetAsString("foo", "b")
ok = False
try:
    ot.ResourceMap.SetAsString("foo", "z")
except Exception:
    ok = True
assert ok
ot.ResourceMap.RemoveKey("foo")
ot.ResourceMap.AddAsString("foo", "a")
assert not ot.ResourceMap.HasStringEnum("foo")



"""
Define a function with a field output: the viscous free fall example
====================================================================
"""

# %%
# In this example, we define a function which has a vector input and a field output.
# This is why we use the :class:`~openturns.PythonPointToFieldFunction` class to create the associated function and propagate the uncertainties through it.
#
# We consider a viscous free fall as explained :ref:`here <use-case-viscous-fall>`.

# %%
# Define the model
# ----------------

# %%
import openturns as ot
import openturns.viewer as otv
from openturns.usecases import viscous_free_fall


# %%
# Load the viscous free fall example.
vff = viscous_free_fall.ViscousFreeFall()
distribution = vff.distribution
model = vff.model

# %%
# Sample trajectories
# -------------------

# %%
# In order to sample trajectories, we use the `getSample` method of the input distribution and apply the field function.

# %%
size = 10
inputSample = distribution.getSample(size)
outputSample = model(inputSample)

# %%
ot.ResourceMap.SetAsUnsignedInteger("Drawable-DefaultPalettePhase", size)

# %%
# Draw some curves.

# %%
graph = outputSample.drawMarginal(0)
graph.setTitle("Viscous free fall: %d trajectories" % (size))
graph.setXTitle(r"$t$")
graph.setYTitle(r"$z$")
view = otv.View(graph)

# %%
# We see that the object first moves up and then falls down.
# Not all objects, however, achieve the same maximum altitude.
# We see that some trajectories reach a higher maximum altitude than others.
# Moreover, at the final time :math:`t_{max}`, one trajectory hits the ground: :math:`z(t_{max})=0` for this trajectory.
otv.View.ShowAll()



#! /usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()


# TimeGrid parameters
n = 101
timeStart = 0.0
timeStep = 0.1
timeGrid = ot.RegularGrid(timeStart, timeStep, n)

# White noise
whiteNoise = ot.WhiteNoise(ot.Uniform(), timeGrid)

# Composite process
process = ot.CompositeProcess(
    ot.ValueFunction(ot.SymbolicFunction("x", "x+2"), timeGrid), whiteNoise
)
# A realization of the process
timeSeries = process.getRealization()
sample = timeSeries.getValues()

# Now we build the factory
factory = ot.BoxCoxFactory()

# Creation of the BoxCoxTransform
myBoxCox = factory.build(timeSeries)

print("myBoxCox (time-series)=", myBoxCox)
print("myBoxCox (sample)     =", factory.build(sample))

# Creation of the BoxCoxTransform using shift
shift = ot.Point(1, 1.0)
myBoxCoxShift, graph = factory.buildWithGraph(timeSeries, shift)

print("myBoxCox with shift (time-series)=", myBoxCoxShift)
print("myBoxCox with shift (sample)     =", factory.build(sample, shift))
print("BoxCox graph (time-series)=", graph)

# with sample
myBoxCoxShift, graph = factory.buildWithGraph(sample, shift)



"""
Draw a survival function
========================
"""

# sphinx_gallery_thumbnail_number = 9
# %%
#
# Introduction
# ------------
#
# The goal of this example is to show how to draw the survival function of a
# sample or a distribution, in linear and logarithmic scales.
#
# Let :math:`X` be a random variable with distribution function :math:`F`:
#
# .. math::
#    F(x) = P(X\leq x)
#
#
# for any :math:`x\in\mathbb{R}`.
# The survival function :math:`S` is:
#
# .. math::
#    S(x) = P(X>x) = 1 - P(X\leq x) = 1 - F(x)
#
#
# for any :math:`x\in\mathbb{R}`.
#
# Let us assume that :math:`\{x_1,...,x_N\}` is a sample from :math:`F`.
#
# Let :math:`\hat{F}_N` be the empirical cumulative distribution function:
#
# .. math::
#    \hat{F}_N(x) = \frac{1}{N} \sum_{i=1}^N \mathbf{1}_{x_i\leq x}
#
#
# for any :math:`x\in\mathbb{R}`.
# Let :math:`\hat{S}_n` be the empirical survival function:
#
# .. math::
#    \hat{S}_N(x) = \frac{1}{N} \sum_{i=1}^N \mathbf{1}_{x_i>x}
#
#
# for any :math:`x\in\mathbb{R}`.
#
# Motivations for the survival function
# -------------------------------------
#
# For many probabilistic models associated with extreme events or lifetime models,
# the survival function has a simpler expression than the distribution function.
#
# * More specifically, several models (e.g., Pareto or Weibull) have a simple
#   expression when we consider the logarithm of the survival function.
#   In this situation, the :math:`(\log(x),\log(S(x)))` plot is often used.
#   For some distributions, this plot is a straight line.
#
# * When we consider probabilities very close to 1 (e.g., with extreme events),
#   a loss of precision can occur when we consider the :math:`1-F(x)` expression
#   with floating point numbers.
#   This loss of significant digits is known as "catastrophic cancellation" in
#   the bibliography and happens when two close floating point numbers are subtracted.
#   This is one of the reasons why we sometimes use directly the survival
#   function instead of the complementary of the distribution.

# %%
# Define a distribution
# ---------------------

# %%
import openturns as ot
import openturns.viewer as otv


# %%
sigma = 1.4
xi = 0.5
u = 0.1
distribution = ot.GeneralizedPareto(sigma, xi, u)

# %%
# Draw the survival of a distribution
# -----------------------------------

# %%
# The `computeCDF` and `computeSurvivalFunction` compute the CDF :math:`F` and survival :math:`S` of a distribution.

# %%
p1 = distribution.computeCDF(10.0)
p1

# %%
p2 = distribution.computeSurvivalFunction(10.0)
p2

# %%
p1 + p2

# %%
# The `drawCDF` and `drawSurvivalFunction` methods allow one to draw the functions :math:`F` and :math:`S`.

# %%
graph = distribution.drawCDF()
graph.setTitle("CDF of a distribution")
view = otv.View(graph)

# %%
graph = distribution.drawSurvivalFunction()
graph.setTitle("Survival function of a distribution")
view = otv.View(graph)

# %%
# In order to get finite bounds for the next graphics, we compute the `xmin`
# and `xmax` bounds from the 0.01 and 0.99 quantiles of the distributions.

# %%
xmin = distribution.computeQuantile(0.01)[0]
xmin

# %%
xmax = distribution.computeQuantile(0.99)[0]
xmax

# %%
# The `drawSurvivalFunction` method also has an option to plot the survival with the X axis in logarithmic scale.

# %%
npoints = 50
logScaleX = True
graph = distribution.drawSurvivalFunction(xmin, xmax, npoints, logScaleX)
graph.setTitle("Survival function of a distribution where X axis is in log scale")
view = otv.View(graph)
# graph

# %%
# In order to get both axes in logarithmic scale, we use the `LOGXY` option of the graph.

# %%
npoints = 50
logScaleX = True
graph = distribution.drawSurvivalFunction(xmin, xmax, npoints, logScaleX)
graph.setLogScale(ot.GraphImplementation.LOGXY)
graph.setTitle(
    "Survival function of a distribution where X and Y axes are in log scale"
)
view = otv.View(graph)
# graph

# %%
# Compute the survival of a sample
# --------------------------------

# %%
# We now generate a sample that we are going to analyze.

# %%
sample = distribution.getSample(1000)

# %%
sample.getMin(), sample.getMax()

# %%
# The `computeEmpiricalCDF` method of a `Sample` computes the empirical CDF.

# %%
p1 = sample.computeEmpiricalCDF([10])
p1

# %%
# Activating the second optional argument allows one to compute the empirical survival function.

# %%
p2 = sample.computeEmpiricalCDF([10], True)
p2

# %%
p1 + p2

# %%
# Draw the survival of a sample
# -----------------------------

# %%
# In order to draw the empirical functions of a `Sample`, we use the :class:`~openturns.FiniteDiscreteDistribution` class.
#
# * The `drawCDF` method plots the CDF.
# * The `drawSurvivalFunction` method plots the survival function.

# %%
userdefined = ot.FiniteDiscreteDistribution(sample)
graph = userdefined.drawCDF()
graph.setTitle("CDF of a sample")
view = otv.View(graph)
# graph

# %%
graph = userdefined.drawSurvivalFunction()
graph.setTitle("Empirical survival function of a sample")
view = otv.View(graph)
# graph

# %%
# As previously, the `drawSurvivalFunction` method of a distribution has an option to set the X axis in logarithmic scale.

# %%
xmin = sample.getMin()[0]
xmax = sample.getMax()[0]
pointNumber = sample.getSize()
logScaleX = True
graph = userdefined.drawSurvivalFunction(xmin, xmax, pointNumber, logScaleX)
graph.setTitle("Empirical survival function of a sample; X axis in log-scale")
view = otv.View(graph)
# graph

# %%
# We obviously have :math:`P(X>X_{max})=0`, where :math:`X_{max}` is the sample maximum.
# This prevents from using the sample maximum and have a logarithmic Y axis at the same time.
# This is why in the following example we restrict the interval where we draw the survival function.

# %%
xmin = sample.getMin()[0]
xmax = sample.getMax()[0] - 1  # To avoid log(0) because P(X>Xmax)=0
pointNumber = sample.getSize()
logScaleX = True
graph = userdefined.drawSurvivalFunction(xmin, xmax, pointNumber, logScaleX)
graph.setLogScale(ot.GraphImplementation.LOGXY)
graph.setTitle("Empirical survival function of a sample; X and Y axes in log-scale")
view = otv.View(graph)
# graph

# %%
# Compare the distribution and the sample with respect to the survival
# --------------------------------------------------------------------

# %%
# In the final example, we compare the distribution and sample survival function in the same graphics.

# %%
xmin = sample.getMin()[0]
xmax = sample.getMax()[0] - 1  # To avoid log(0) because P(X>Xmax)=0
npoints = 50
logScaleX = True
graph = userdefined.drawSurvivalFunction(xmin, xmax, pointNumber, logScaleX)
graph.setLogScale(ot.GraphImplementation.LOGXY)
graph.setLegends(["Sample"])
graphDistribution = distribution.drawSurvivalFunction(xmin, xmax, npoints, logScaleX)
graphDistribution.setLegends(["GPD"])
graph.add(graphDistribution)
graph.setLegendPosition("upper right")
graph.setTitle("GPD against the sample - n=%d" % (sample.getSize()))
view = otv.View(graph)


# %%
# Show all the graphs.
view.ShowAll()



#! /usr/bin/env python

import openturns as ot
import openturns.testing as ott
from openturns.usecases import ishigami_function

ot.TESTPREAMBLE()
ot.RandomGenerator.SetSeed(0)

# Ishigami use-case
ishigami = ishigami_function.IshigamiModel()
distX = ishigami.distribution

# Get a sample of it
size = 100
X = distX.getSample(size)


# The Ishigami model
modelIshigami = ishigami.model
modelIshigami.setParameter([5, 0.1])

# Output
Y = modelIshigami(X)

# Using the same covariance model for each marginal
Cov1 = ot.SquaredExponential(1)

# Output covariance model
Cov2 = ot.SquaredExponential(1)

# Set output covariance scale
Cov2.setScale(Y.computeStandardDeviation())

# This is the GSA-type estimator: weight is 1.
W = ot.Point(size, 1.0)

# Using a biased estimator
estimatorTypeV = ot.HSICVStat()

# Loop over marginals
hsicIndexRef = [0.02331323, 0.00205350, 0.00791711]
for i in range(3):
    test = X.getMarginal(i)
    # Set input covariance scale
    Cov1.setScale(test.computeStandardDeviation())
    hsicIndex = estimatorTypeV.computeHSICIndex(
        Cov1.discretize(test), Cov2.discretize(Y), W
    )
    ott.assert_almost_equal(hsicIndex, hsicIndexRef[i])

# Using an unbiased estimator
estimatorTypeU = ot.HSICUStat()

# Loop over marginals
hsicIndexRef = [0.02228377, 0.00025668, 0.00599247]
for i in range(3):
    test = X.getMarginal(i)
    # Set input covariance scale
    Cov1.setScale(test.computeStandardDeviation())
    hsicIndex = estimatorTypeU.computeHSICIndex(
        Cov1.discretize(test), Cov2.discretize(Y), W
    )
    ott.assert_almost_equal(hsicIndex, hsicIndexRef[i])



#! /usr/bin/env python

import openturns as ot
import openturns.testing as ott

ot.TESTPREAMBLE()

matrix1 = ot.CovarianceMatrix(2)
print("matrix1 (default)=" + repr(matrix1))
matrix1[0, 0] = 1.0
matrix1[1, 0] = 0.5
matrix1[1, 1] = 1.0
print("matrix1 (initialized)=" + repr(matrix1))

pt = ot.Point()
pt.add(5.0)
pt.add(0.0)
print("pt=", repr(pt))

result = ot.Point()
result = matrix1.solveLinearSystem(pt)
print("result=" + repr(result))

determinant = matrix1.computeDeterminant()
print("determinant=%.6f" % determinant)

ev = ot.ScalarCollection(2)
ev = matrix1.computeEigenValues()
print("ev=" + repr(ev))

if matrix1.isPositiveDefinite():
    isSPD = "true"
else:
    isSPD = "false"
print("isSPD=", isSPD)

matrix2 = matrix1.computeCholesky()
print("matrix2=" + repr(matrix2))

b = ot.Matrix(2, 3)
b[0, 0] = 5.0
b[1, 0] = 0.0
b[0, 1] = 10.0
b[1, 1] = 1.0
b[0, 2] = 15.0
b[1, 2] = 2.0
result2 = matrix1.solveLinearSystem(b)
print("result2=" + repr(result2))

matrix3 = ot.CovarianceMatrix(3)
matrix3[1, 0] = float("nan")
try:
    print("ev=", matrix3.computeSingularValues())
except Exception:
    print("ok")

# from SymmetricMatrix
sym = ot.SymmetricMatrix(3)
sym[0, 0] = 1.0e-02
sym[1, 1] = 1.0e-02
sym[2, 2] = 1.0e-02
sym[0, 1] = 7.0e-04
ot.CovarianceMatrix(sym)
print("ok")

# regularized cholesky
A = ot.CovarianceMatrix([[1.0] * 10] * 10)
L = A.computeRegularizedCholesky()
B = L * L.transpose()
ott.assert_almost_equal(A, B)



#! /usr/bin/env python

import openturns as ot


ot.TESTPREAMBLE()


# We create a numerical math function
myFunction = ot.SymbolicFunction(["x0", "x1"], ["-(6+x0^2-x1)"])

dim = myFunction.getInputDimension()
# We create a normal distribution point of dimension 1
mean = [5.0, 2.1]
sigma = [3.3, 3.0]
R = ot.IdentityMatrix(dim)

#
testDistributions = ot.DistributionCollection(2)
testDistributions[0] = ot.Normal(mean, sigma, R)
marginals = ot.DistributionCollection(2)
marginals[0] = testDistributions[0].getMarginal(0)
marginals[1] = testDistributions[0].getMarginal(1)
testDistributions[1] = ot.JointDistribution(marginals, ot.NormalCopula(R))
for i in range(1):
    myDistribution = testDistributions[i]
    myDistribution.setDescription(["Marginal 1", "Marginal 2"])

    # We create a 'usual' RandomVector from the Distribution
    vect = ot.RandomVector(myDistribution)

    # We create a composite random vector
    output = ot.CompositeRandomVector(myFunction, vect)
    output.setDescription(["Interest Variable 1"])

    # We create an Event from this RandomVector
    myEvent = ot.ThresholdEvent(output, ot.Greater(), 0.0)

    # We create a NearestPoint algorithm
    myCobyla = ot.Cobyla()
    myCobyla.setStartingPoint(mean)
    myCobyla.setMaximumCallsNumber(200)
    myCobyla.setMaximumAbsoluteError(1.0e-10)
    myCobyla.setMaximumRelativeError(1.0e-10)
    myCobyla.setMaximumResidualError(1.0e-10)
    myCobyla.setMaximumConstraintError(1.0e-10)
    print("myCobyla=", myCobyla)

    # We create a FORM algorithm
    # The first parameter is an OptimizationAlgorithm
    # The second parameter is an event
    # The third parameter is a starting point for the design point research
    myAlgo = ot.FORM(myCobyla, myEvent)

    print("FORM=", myAlgo)

    # Perform the simulation
    myAlgo.run()

    # Stream out the result
    result = ot.FORMResult(myAlgo.getResult())
    print("importance factors=", result.getImportanceFactors())
    print("Hasofer reliability index=%.6f" % result.getHasoferReliabilityIndex())
    print("result=", result)

    # Hasofer Reliability Index Sensitivity
    hasoferReliabilityIndexSensitivity = result.getHasoferReliabilityIndexSensitivity()
    print(
        "hasoferReliabilityIndexSensitivity = ",
        repr(hasoferReliabilityIndexSensitivity),
    )

    # Event Probability Sensitivity
    eventProbabilitySensitivity = result.getEventProbabilitySensitivity()
    print("eventProbabilitySensitivity = ", repr(eventProbabilitySensitivity))
